#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Create a Lithology object with different properties."""

import numpy as np
import xarray as xr
from scipy.interpolate import interp1d

from landlab import Component
from landlab.layers import EventLayers, MaterialLayers


class Lithology(Component):

    """Create a Lithology object.

    A Lithology is a three dimentional representation of material operated on
    by landlab components. Material can be removed through erosion or added to
    through deposition. Rock types can have multiple attributes (e.g. age,
    erodability or other parameter values, etc).

    If the tracked properties are model grid fields, they will be updated to
    the surface values of the Lithology. If the properties are not grid fields
    then at-node grid fields will be created with their names. Lithology and
    its derived versions will make a at-node grid field called `rock_type__id`
    to store the rock type id.

    Lithology was designed to be used on its own and to be inherited from and
    improved. Currently one other Lithology variant exists: LithoLayers
    which makes it easy to specify parallel layers of rock with generic layer
    geometries.

    It is constructed by specifying a series of thicknesses and a series of
    rock type IDs. Thicknesses and IDs are both specified in order of closest
    to the surface to furthest from the surface. Thicknesses can either be a
    single value (corresponding to a layer of uniform thickness) or a number-of
    -nodes length array (corresponding to a non-uniform layer).

    Additionally, an attribute dictionary specifies the properties of each
    rock type. This dictionary is expected to have the form of:

    .. code-block:: python

        attrs = {'K_sp': {1: 0.001,
                          2: 0.0001},
                 'D': {1: 0.01,
                       2: 0.001}}

    Where ``'K_sp'`` and ``'D'`` are properties to track, and ``1`` and ``2``
    are rock type IDs. The rock type IDs can be any type that is valid as a
    python dictionary key.
    """

    _name = "Lithology"

    _cite_as = """@article{barnhart2018lithology,
                    title = "Lithology: A Landlab submodule for spatially variable rock properties",
                    journal = "Journal of Open Source Software",
                    volume = "",
                    pages = "",
                    year = "2018",
                    doi = "10.21105/joss.00979",
                    author = "Katherine R. Barnhart and Eric Hutton and Nicole M. Gasparini and Gregory E. Tucker",
                    }"""

    def __init__(self, grid, thicknesses, ids, attrs, layer_type="MaterialLayers"):
        """Create a new instance of Lithology.

        Parameters
        ----------
        grid : Landlab ModelGrid
        thicknesses : ndarray of shape `(n_layers, )` or `(n_layers, n_nodes)`
            Values of layer thicknesses from surface to depth. Layers do not
            have to have constant thickness. Layer thickness can be zero,
            though the entirety of Lithology must have non-zero thickness.
        ids : ndarray of shape `(n_layers, )` or `(n_layers, n_nodes)`
            Values of rock type IDs corresponding to each layer specified in
            **thicknesses**. A single layer may have multiple rock types if
            specified by the user.
        attrs : dict
            Rock type property dictionary. See class docstring for example of
            required format.
        layer_type : str, optional
            Type of Landlab layers object used to store the layers. If
            MaterialLayers (default) is specified, then erosion removes material
            and does not create a layer of thickness zero. If EventLayers is
            used, then erosion removes material and creates layers of thickness
            zero. Thus, EventLayers may be appropriate if the user is interested
            in chronostratigraphy.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')

        Create a Lithology with uniform thicknesses that alternates between
        layers of type 1 and type 2 rock.

        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)

        After creating a Lithology, the model grid will have an at-node grid
        field set to the surface values of 'K_sp'.

        >>> mg.at_node['K_sp']
        array([ 0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,
                0.001])

        The surface values are also properties of the Lithology.

        >>> lith['K_sp']
        array([ 0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,
                0.001])

        We can access information about the Lithology like the total thickness
        or layer thicknesses.

        >>> lith.thickness
        array([ 8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.])
        >>> lith.dz
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.]])

        This might look confusing -- that the layers are in reverse order, but
        it is OK. The last layers in the Lithology are those that are closest
        to the surface.

        The layers don't all have to have the same thickness as in the prior
        example. If the layers have non-uniform thickness, then they must be
        specified in an array of shape `(n_layer, n_nodes)`. In this case, the
        layer IDs must be specified in either an array of `(n_layer)` or
        `(n_layer, n_nodes)`.

        Here we make a layer that gets thicker as a function of the x value of
        the model grid.

        >>> layer_pattern = (0.5 * mg.x_of_node) + 1.0
        >>> thicknesses = [1*layer_pattern, 2*layer_pattern, 4*layer_pattern]
        >>> ids = [1, 2, 1]
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.thickness
        array([  7. ,  10.5,  14. ,   7. ,  10.5,  14. ,   7. ,  10.5,  14. ])
        >>> lith.dz
        array([[ 4. ,  6. ,  8. ,  4. ,  6. ,  8. ,  4. ,  6. ,  8. ],
               [ 2. ,  3. ,  4. ,  2. ,  3. ,  4. ,  2. ,  3. ,  4. ],
               [ 1. ,  1.5,  2. ,  1. ,  1.5,  2. ,  1. ,  1.5,  2. ]])
        """
        # save reference to the grid and the last time steps's elevation.
        self._grid = grid

        try:
            self.last_elevation = self._grid["node"]["topographic__elevation"][:].copy()
        except KeyError:
            msg = (
                "Lithology requires that topographic__elevation already "
                "exists as an at-node field."
            )
            raise ValueError(msg)

        # save inital information about thicknesses, layers, attributes, and ids.
        self._init_thicknesses = np.asarray(thicknesses)
        self._attrs = attrs
        self._number_of_init_layers = self._init_thicknesses.shape[0]
        self._properties = list(attrs.keys())
        self._rock_id_name = "rock_type__id"
        # assert that thicknesses and ids are correct and consistent shapes

        # if thickness is a 2d array.
        if self._init_thicknesses.ndim == 2:
            # assert that the 2nd dimension is the same as the number of nodes.
            if self._init_thicknesses.shape[1] != self._grid.number_of_nodes:
                msg = (
                    "Thicknesses provided to Lithology are ",
                    "inconsistent with the ModelGrid.",
                )
                raise ValueError(msg)

            # if IDs is a 2d array assert that it is the same size as thicknesses
            if np.asarray(ids).ndim == 2:
                if self._init_thicknesses.shape != np.asarray(ids).shape:
                    msg = (
                        "Thicknesses and IDs provided to Lithology are ",
                        "inconsistent with each other.",
                    )
                    raise ValueError(msg)
                # if tests pass set value of IDs.
                self._layer_ids = np.asarray(ids)

            # if IDS is a 1d array
            elif np.asarray(ids).ndim == 1:
                if np.asarray(ids).size != self._number_of_init_layers:
                    msg = (
                        "Number of IDs provided to Lithology is ",
                        "inconsistent with number of layers provided in "
                        "thicknesses.",
                    )
                    raise ValueError(msg)
                # if tests pass, broadcast ids to correct shape.
                self._layer_ids = np.broadcast_to(
                    np.atleast_2d(np.asarray(ids)).T, self._init_thicknesses.shape
                )

            else:
                msg = (
                    "IDs must be of shape `(n_layers, )` or `(n_layers, "
                    "n_nodes)`. Passed array has more than 2 dimensions."
                )
                raise ValueError(msg)

        elif self._init_thicknesses.ndim == 1:
            if self._init_thicknesses.shape != np.asarray(ids).shape:
                msg = (
                    "Thicknesses and IDs provided to Lithology are ",
                    "inconsistent with each other.",
                )
                raise ValueError(msg)
            self._layer_ids = np.asarray(ids)
        else:
            msg = (
                "Thicknesses must be of shape `(n_layers, )` or `(n_layers, "
                "n_nodes)`. Passed array has more than 2 dimensions."
            )
            raise ValueError(msg)

        # assert that attrs are pointing to fields (or create them)
        for at in self._properties:
            if at not in grid.at_node:
                self._grid.add_empty("node", at)

        # add a field for the rock type id
        if self._rock_id_name not in self._grid.at_node:
            self._grid.add_empty("node", self._rock_id_name)

        # verify that all IDs have attributes.
        self._check_property_dictionary()

        # create a EventLayers instance
        if layer_type == "EventLayers":
            self._layers = EventLayers(
                grid.number_of_nodes, self._number_of_init_layers
            )
        elif layer_type == "MaterialLayers":
            self._layers = MaterialLayers(
                grid.number_of_nodes, self._number_of_init_layers
            )
        else:
            raise ValueError(("Lithology passed an invalid option for " "layer type."))

        # From bottom to top, add layers to the Lithology with attributes.
        for i in range(self._number_of_init_layers - 1, -1, -1):
            try:
                self.add_layer(self._init_thicknesses[i, :], self._layer_ids[i, :])
            except IndexError:
                self.add_layer(self._init_thicknesses[i], self._layer_ids[i])

    def __getitem__(self, name):
        return self._get_surface_values(name)

    @property
    def ids(self):
        """Rock type IDs used by Lithology."""
        return list(self._ids)

    @property
    def tracked_properties(self):
        """Properties tracked by Lithology.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.tracked_properties
        ['K_sp']
        """
        self._properties.sort()
        return self._properties

    @property
    def properties(self):
        """Properties dictionary used by Lithology.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.properties
        {'K_sp': {1: 0.001, 2: 0.0001}}
        """
        return self._attrs

    @property
    def thickness(self):
        """Total thickness of the Lithology at each node.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.thickness
        array([ 8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.])
        """
        return self._layers.thickness

    @property
    def dz(self):
        """Thickness of each layer in the Lithology at each node.

        The thickness of each layer in the Lithology as an array of shape
        `(number_of_layers, number_of_nodes)`.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.dz
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.]])
        """
        return self._layers.dz

    @property
    def z_bottom(self):
        """Thickness from the surface to the bottom of each layer in Lithology.

        Thickness from the topographic surface to the bottom of each layer as
        an array of shape `(number_of_layers, number_of_nodes)`.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.z_bottom
        array([[ 8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.],
               [ 7.,  7.,  7.,  7.,  7.,  7.,  7.,  7.,  7.],
               [ 3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.]])
        """
        thick = np.broadcast_to(self._layers.thickness, self._layers.z.shape)
        return thick - self._layers.z + self._layers.dz

    @property
    def z_top(self):
        """Thickness from the surface to the top of each layer in Lithology.

        Thickness from the topographic surface to the top of each layer as
        an array of shape `(number_of_layers, number_of_nodes)`.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.z_top
        array([[ 7.,  7.,  7.,  7.,  7.,  7.,  7.,  7.,  7.],
               [ 3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.,  3.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.]])
        """
        thick = np.broadcast_to(self._layers.thickness, self._layers.z.shape)
        return thick - self._layers.z

    def _check_property_dictionary(self):
        """Check compatibility of Lithology and property dictionary."""
        ids = []
        for at in self._properties:
            ids.extend(self._attrs[at].keys())
        self._ids = frozenset(np.unique(ids))

        for at in self._properties:
            for i in self._ids:
                if i not in self._attrs[at]:
                    msg = (
                        "A rock type with ID value " + str(i) + "was "
                        "specified in Lithology. No value "
                        "for this ID was provided in property " + at + "."
                    )
                    raise ValueError(msg)

    def _update_surface_values(self):
        """Update Lithology surface values"""
        # Update surface values for each attribute.
        self._grid["node"][self._rock_id_name][:] = self._surface_rock_type
        for at in self._properties:
            self._grid["node"][at][:] = self[at]

    def add_layer(self, thickness, rock_id=None):
        """Add a new layer to Lithology.

        Parameters
        ----------
        thickness : float or `(n_nodes,)` array
            Positive values deposit material on to Lithology while negative
            values erode Lithology.
        rock_id : single value or `n_nodes` long itterable, optional if only erosion occurs
            Rock type ID for new deposits. Can be single value or an number-
            of-nodes array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]

        We can instantiate Lithology with rock type properties we know we will
        use in the future.

        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001,
        ...                   3: 0.01}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)

        Add a layer of thickness 3 and rock type 3.

        >>> lith.add_layer(3, rock_id=3)

        The value of K_sp at node is now updated to the value of rock type 3

        >>> mg.at_node['K_sp']
        array([ 0.01,  0.01,  0.01,  0.01,  0.01,  0.01,  0.01,  0.01,  0.01])

        A negative value will erode. We can also pass a `(n_nodes,) long array
        to erode unevenly. If all parts of the layer erode, then no rock_id
        needs to be passed.

        >>> erosion_amount = [-2., -2., -2., -4., -4., -4., -6., -6., -6.]
        >>> lith.add_layer(erosion_amount)
        >>> mg.at_node['K_sp']
        array([ 0.01  ,  0.01  ,  0.01  ,  0.0001,  0.0001,  0.0001,  0.001 ,
                0.001 ,  0.001 ])

        Now different layers are exposed at the surface and the value of K_sp
        is spatially variable.
        """
        thickness = np.array(thickness)

        # verify that Lithology will still have thickness after change
        if np.any((self._layers.thickness + thickness) <= 0):
            msg = (
                "add_layer will result in Lithology having a thickness of "
                "zero at at least one node."
            )
            raise ValueError(msg)

        # verify that rock type added exists.
        try:
            all_ids_present = self._ids.issuperset(rock_id)
            new_ids = rock_id
        except TypeError:
            all_ids_present = self._ids.issuperset([rock_id])
            new_ids = [rock_id]

        if not all_ids_present:

            missing_ids = set(new_ids).difference(self._ids)

            if np.any(thickness > 0):
                msg = (
                    "Lithology add_layer was given a rock type id that does "
                    "not yet exist and will need to deposit. Use a valid "
                    "rock type or add_rock_type. " + str(missing_ids)
                )
                raise ValueError(msg)

        # add_rock_type
        if rock_id is not None:
            # add layer
            attributes = {self._rock_id_name: rock_id}
            self._layers.add(thickness, **attributes)
        else:
            self._layers.add(thickness)

        # update surface rock type
        self._surface_rock_type = self._layers.get_surface_values(self._rock_id_name)

        # update surface values
        self._update_surface_values()

    def add_property(self, attrs):
        """Add new property to Lithology

        Parameters
        ----------
        attrs : dict
            Rock attribute dictionary for the new property(s).

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.add_property({'D': {1: 0.03,
        ...                        2: 0.004}})
        >>> lith.tracked_properties
        ['D', 'K_sp']
        >>> mg.at_node['D']
        array([ 0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03])
        """
        for at in attrs:
            if at in self._properties:
                msg = (
                    "add_property is trying to add an existing "
                    "attribute, this is not permitted. " + str(at)
                )
                raise ValueError(msg)

            new_rids = attrs[at].keys()
            for rid in new_rids:
                if rid not in self._ids:
                    msg = (
                        "add_property has an attribute(" + str(at) + ")"
                        " for rock type " + str(rid) + " that no other rock "
                        " type has. This is not permitted."
                    )
                    raise ValueError(msg)

            for rid in self._ids:
                if rid not in new_rids:
                    msg = (
                        "add_property needs a value for id " + str(rid) + ""
                        " and attribute " + str(at) + "."
                    )
                    raise ValueError(msg)

        for at in attrs:
            if at not in self._grid.at_node:
                self._grid.add_empty("node", at)
            self._attrs[at] = attrs[at]
            self._properties.append(at)

        # update surface values
        self._update_surface_values()

    def add_rock_type(self, attrs):
        """Add rock type to Lithology.

        Parameters
        ----------
        attrs : dict
            Rock attribute dictionary for the new rock type(s).

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.add_rock_type({'K_sp': {4: 0.03,
        ...                            6: 0.004}})
        >>> lith.ids
        [1, 2, 4, 6]
        >>> lith.properties
        {'K_sp': {1: 0.001, 2: 0.0001, 4: 0.03, 6: 0.004}}

        """
        # Check that the new rock type has all existing attributes
        for at in self._properties:
            if at not in attrs:
                msg = "The new rock type is missing attribute " + str(at) + "."
                raise ValueError(msg)
        # And no new attributes
        for at in attrs:
            if at not in self._properties:
                msg = (
                    "The new rock type has an attribute (e" + str(at) + ") "
                    "that no other rock type has. This is not permitted."
                )
                raise ValueError(msg)

        new_ids = []
        for at in attrs:
            att_dict = attrs[at]
            rids = att_dict.keys()
            for rid in rids:
                if rid in self._layer_ids:
                    msg = (
                        "Rock type ID " + str(rid) + " for attribute "
                        "" + str(at) + " has already been added. This is "
                        "not allowed"
                    )
                    raise ValueError(msg)
                else:
                    new_ids.append(rid)
                    self._attrs[at][rid] = att_dict[rid]
        self._ids = self._ids.union(new_ids)

        # update surface values
        self._update_surface_values()

    def update_rock_properties(self, at, rock_id, value):
        """Update rock type attribute.

        Paramters
        ---------
        at : str
            Attribute name
        rock_id : value
            Rock type ID
        value : value
            New value for rock type attribute

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)

        >>> mg.at_node['K_sp']
        array([ 0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,  0.001,
                0.001])

        >>> lith.update_rock_properties('K_sp', 1, 0.03)

        >>> mg.at_node['K_sp']
        array([ 0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03,  0.03])

        """
        if at not in self._properties:
            msg = (
                "Lithology cannot update the value of " + str(at) + "as "
                "this attribute does not exist."
            )
            raise ValueError(msg)

        if not self._ids.issuperset([rock_id]):
            msg = (
                "Lithology cannot update the value of rock type "
                "" + str(rock_id) + "for attribute " + str(at) + " as "
                "this rock type is not yet defined."
            )
            raise ValueError(msg)

        # set the value in the attribute dictionary
        self._attrs[at][rock_id] = value

        # update surface values
        self._update_surface_values()

    def _get_surface_values(self, at):
        """Get surface values for attribute."""
        return np.array(list(map(self._attrs[at].get, self._surface_rock_type)))

    def rock_cube_to_xarray(self, depths):
        """Construct a 3D rock cube of rock type ID as an xarray dataset.

        Create an xarray dataset in (x, y, z) that shows the rock type with
        depth relative to the current topographic surface.

        Here the z dimension is depth relative to the current topographic
        surface, NOT depth relative to an absolute datum.

        Note also that when this method is called, it will construct the current
        values of lithology with depth, NOT the initial values.

        Paramters
        ---------
        depths : array

        Returns
        -------
        ds : xarray dataset
        """
        depths = np.asarray(depths)
        rock_type = self._layers[self._rock_id_name]
        rock_cube = np.empty((depths.size, self._grid.shape[0], self._grid.shape[1]))

        # at each node point, interpolate between ztop/bottomo correct.y
        for sid in range(self._layers.number_of_stacks):
            coord = np.unravel_index(sid, (self._grid.shape[0], self._grid.shape[1]))
            real_layers = self.dz[:, sid] > 0
            f = interp1d(
                np.flipud(self.z_top[real_layers, sid]),
                np.flipud(rock_type[real_layers, sid]),
                kind="previous",
            )
            vals = f(depths)
            rock_cube[:, coord[0], coord[1]] = vals

        ds = xr.Dataset(
            data_vars={
                "rock_type__id": (
                    ("z", "y", "x"),
                    rock_cube,
                    {"units": "-", "long_name": "Rock Type ID Code"},
                )
            },
            coords={
                "x": (
                    ("x"),
                    self._grid.x_of_node.reshape(self._grid.shape)[0, :],
                    {"units": "meters"},
                ),
                "y": (
                    ("y"),
                    self._grid.y_of_node.reshape(self._grid.shape)[:, 1],
                    {"units": "meters"},
                ),
                "z": (
                    ("z"),
                    depths,
                    {"units": "meters", "long_name": "Depth Below Topographic Surface"},
                ),
            },
        )

        return ds

    def run_one_step(self, dz_advection=0, rock_id=None):
        """Update Lithology.

        The ``run_one_step`` method calculates elevation change of the
        Lithology surface (taking into account any advection due to external
        processes) and then either deposits or erodes based on elevation
        change.

        Parameters
        ----------
        dz_advection : float or `(n_nodes, ) shape array, optional
            Change in rock elevation due to advection by some external process.
        rock_id : value or `(n_nodes, ) shape array, optional
            Rock type id for new material if deposited.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import Lithology
        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_ones('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs)
        >>> lith.dz
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.]])
        >>> lith.thickness
        array([ 8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.])

        If we erode the surface, and then update Lithology, the thickness will
        change.

        >>> z -= 0.5
        >>> lith.run_one_step(dz_advection=0)
        >>> lith.thickness
        array([ 7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5])

        The default of Lithology is to use MaterialLayers from the Landlab
        layers submodule. This means that when we erode, we will remove a layer
        from the layers datastructure if it has no material anywhere.

        >>> lith.dz
        array([[ 1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ],
               [ 4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ],
               [ 2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ],
               [ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5]])

        We can see the value of the rock type at the surface.

        >>> mg.at_node['rock_type__id']
        array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

        If you deposit, a valid rock_id must be provided. If the rock type
        is the same as the current surface value everywhere, then the layers
        will be combined.

        >>> z += 1.5
        >>> lith.run_one_step(dz_advection=0, rock_id = 1)
        >>> lith.thickness
        array([ 9.,  9.,  9.,  9.,  9.,  9.,  9.,  9.,  9.])

        >>> lith.dz
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.]])

        This contrasts with the behavior of Lithology if we use EventLayers.
        Next we repeat this example with EventLayers. Note that no matter which
        method you use, the values of the model grid fields will be the same.
        These two methods differ only in the details of the data structure they
        use to store the layer information.

        >>> mg = RasterModelGrid((3, 3))
        >>> z = mg.add_ones('node', 'topographic__elevation')
        >>> thicknesses = [1, 2, 4, 1]
        >>> ids = [1, 2, 1, 2]
        >>> attrs = {'K_sp': {1: 0.001,
        ...                   2: 0.0001}}
        >>> lith = Lithology(mg, thicknesses, ids, attrs,
        ...                  layer_type='EventLayers')
        >>> lith.dz
        array([[ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],
               [ 4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.,  4.],
               [ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.],
               [ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.]])
        >>> lith.thickness
        array([ 8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.,  8.])

        If we erode the surface, and then update Lithology, the thickness
        will change. However, with EventLayers, the ``lith.dz`` structure
        will be different. It will have a layer with thickness zero that
        represents the event of erosion.

        >>> z -= 0.5
        >>> lith.run_one_step(dz_advection=0)
        >>> lith.thickness
        array([ 7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5,  7.5])
        >>> lith.dz
        array([[ 1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ],
               [ 4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ],
               [ 2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ],
               [ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5],
               [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ]])

        We can see the value of the rock type at the surface. As expected,
        it is just the same as if we used MaterialLayers.

        >>> mg.at_node['rock_type__id']
        array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.])

        If you deposit, a valid rock_id must be provided. Unlike
        MaterialLayers, these two layers will not be combined, even if they
        have the same properties.

        >>> z += 1.5
        >>> lith.run_one_step(dz_advection=0, rock_id = 1)
        >>> lith.thickness
        array([ 9.,  9.,  9.,  9.,  9.,  9.,  9.,  9.,  9.])

        >>> lith.dz
        array([[ 1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ],
               [ 4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ,  4. ],
               [ 2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ,  2. ],
               [ 0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5],
               [ 0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ,  0. ],
               [ 1.5,  1.5,  1.5,  1.5,  1.5,  1.5,  1.5,  1.5,  1.5]])
        """
        # calculate amount of erosion
        elevation_change = self._grid["node"]["topographic__elevation"] - (
            self.last_elevation + dz_advection
        )

        # add layer
        self.add_layer(elevation_change, rock_id=rock_id)

        # update the last elevation.
        self.last_elevation = self._grid["node"]["topographic__elevation"][:].copy()



from .fastscape_stream_power import FastscapeEroder
from .sed_flux_dep_incision import SedDepEroder
from .stream_power import StreamPowerEroder
from .stream_power_smooth_threshold import StreamPowerSmoothThresholdEroder

__all__ = [
    "StreamPowerEroder",
    "FastscapeEroder",
    "SedDepEroder",
    "StreamPowerSmoothThresholdEroder",
]



from .flow_accum_bw import (
    find_drainage_area_and_discharge,
    flow_accumulation,
    make_ordered_node_array,
)
from .flow_accumulator import FlowAccumulator
from .lossy_flow_accumulator import LossyFlowAccumulator

__all__ = [
    "FlowAccumulator",
    "LossyFlowAccumulator",
    "make_ordered_node_array",
    "find_drainage_area_and_discharge",
    "flow_accumulation",
]



"""
Unit tests for landlab.components.landslides.landslide_probability
"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from landlab import RasterModelGrid
from landlab.components import LandslideProbability

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(ls_prob):
    """Testing if the name is right.
    """
    assert ls_prob.name == "Landslide Probability"


def test_input_var_names(ls_prob):
    """Testing if the input_var_names outputs the right list.
    """
    assert sorted(ls_prob.input_var_names) == [
        "soil__density",
        "soil__internal_friction_angle",
        "soil__maximum_total_cohesion",
        "soil__minimum_total_cohesion",
        "soil__mode_total_cohesion",
        "soil__saturated_hydraulic_conductivity",
        "soil__thickness",
        "soil__transmissivity",
        "topographic__slope",
        "topographic__specific_contributing_area",
    ]


def test_output_var_names(ls_prob):
    """Testing if output_var_names outputs the right list.
    """
    assert sorted(ls_prob.output_var_names) == [
        "landslide__probability_of_failure",
        "soil__mean_relative_wetness",
        "soil__probability_of_saturation",
    ]


def test_var_units(ls_prob):
    """Testing if units are right.
    """
    assert set(ls_prob.input_var_names) | set(ls_prob.output_var_names), set(
        dict(ls_prob.units).keys()
    )

    assert ls_prob.var_units("topographic__specific_contributing_area") == "m"
    assert ls_prob.var_units("topographic__slope") == "tan theta"
    assert ls_prob.var_units("soil__transmissivity") == "m2/day"
    assert ls_prob.var_units("soil__saturated_hydraulic_conductivity") == "m/day"
    assert ls_prob.var_units("soil__mode_total_cohesion") == "Pa or kg/m-s2"
    assert ls_prob.var_units("soil__minimum_total_cohesion") == "Pa or kg/m-s2"
    assert ls_prob.var_units("soil__maximum_total_cohesion") == "Pa or kg/m-s2"
    assert ls_prob.var_units("soil__internal_friction_angle") == "degrees"
    assert ls_prob.var_units("soil__density") == "kg/m3"
    assert ls_prob.var_units("soil__thickness") == "m"
    assert ls_prob.var_units("soil__mean_relative_wetness") == "None"
    assert ls_prob.var_units("landslide__probability_of_failure") == "None"
    assert ls_prob.var_units("soil__probability_of_saturation") == "None"


def test_grid_shape(ls_prob):
    """Testing if the grid shape matches the inputs.
    """
    assert ls_prob.grid.number_of_node_rows == _SHAPE[0]
    assert ls_prob.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(ls_prob):
    """Testing if x extent is right.
    """
    assert ls_prob.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(ls_prob):
    """Testing if y extent is right.
    """
    assert ls_prob.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(ls_prob):
    """Testing if the right field is called.
    """
    for name in ls_prob.grid["node"]:
        field = ls_prob.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            ls_prob.grid.number_of_node_rows * ls_prob.grid.number_of_node_columns,
        )

    with pytest.raises(KeyError):
        ls_prob.grid["not_a_var_name"]


def test_field_initialized_to_zero(ls_prob):
    """Testing if the fields are initialized with zeros.
    """
    for name in ls_prob.grid["node"]:
        field = ls_prob.grid["node"][name]
        assert_array_almost_equal(field, np.zeros(ls_prob.grid.number_of_nodes))


def test_calculate_landslide_probability_uniform_method():
    """Testing the main method 'calculate_landslide_probability()' with
    'uniform' method.
    """
    grid_1 = RasterModelGrid((5, 4), xy_spacing=(0.2, 0.2))
    gridnum = grid_1.number_of_nodes
    np.random.seed(seed=5)
    grid_1.at_node["topographic__slope"] = np.random.rand(gridnum)
    scatter_dat = np.random.randint(1, 10, gridnum)
    grid_1.at_node["topographic__specific_contributing_area"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_1.at_node["soil__transmissivity"] = np.sort(
        np.random.randint(5, 20, gridnum), -1
    )
    grid_1.at_node["soil__mode_total_cohesion"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_1.at_node["soil__minimum_total_cohesion"] = (
        grid_1.at_node["soil__mode_total_cohesion"] - scatter_dat
    )
    grid_1.at_node["soil__maximum_total_cohesion"] = (
        grid_1.at_node["soil__mode_total_cohesion"] + scatter_dat
    )
    grid_1.at_node["soil__internal_friction_angle"] = np.sort(
        np.random.randint(26, 37, gridnum)
    )
    grid_1.at_node["soil__thickness"] = np.sort(np.random.randint(1, 10, gridnum))
    grid_1.at_node["soil__density"] = 2000.0 * np.ones(gridnum)

    ls_prob_uniform = LandslideProbability(
        grid_1,
        number_of_iterations=10,
        groundwater__recharge_distribution="uniform",
        groundwater__recharge_min_value=20.0,
        groundwater__recharge_max_value=120.0,
        seed=5,
    )
    ls_prob_uniform.calculate_landslide_probability()
    np.testing.assert_almost_equal(
        grid_1.at_node["landslide__probability_of_failure"][5], 1.0
    )
    np.testing.assert_almost_equal(
        grid_1.at_node["landslide__probability_of_failure"][9], 0.0
    )


def test_calculate_landslide_probability_lognormal_method():
    """Testing the main method 'calculate_landslide_probability()' with
    'lognormal' method.
    """
    grid_2 = RasterModelGrid((5, 4), xy_spacing=(0.2, 0.2))
    gridnum = grid_2.number_of_nodes
    np.random.seed(seed=6)
    grid_2.at_node["topographic__slope"] = np.random.rand(gridnum)
    scatter_dat = np.random.randint(1, 10, gridnum)
    grid_2.at_node["topographic__specific_contributing_area"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_2.at_node["soil__transmissivity"] = np.sort(
        np.random.randint(5, 20, gridnum), -1
    )
    grid_2.at_node["soil__mode_total_cohesion"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_2.at_node["soil__minimum_total_cohesion"] = (
        grid_2.at_node["soil__mode_total_cohesion"] - scatter_dat
    )
    grid_2.at_node["soil__maximum_total_cohesion"] = (
        grid_2.at_node["soil__mode_total_cohesion"] + scatter_dat
    )
    grid_2.at_node["soil__internal_friction_angle"] = np.sort(
        np.random.randint(26, 37, gridnum)
    )
    grid_2.at_node["soil__thickness"] = np.sort(np.random.randint(1, 10, gridnum))
    grid_2.at_node["soil__density"] = 2000.0 * np.ones(gridnum)

    ls_prob_lognormal = LandslideProbability(
        grid_2,
        number_of_iterations=10,
        groundwater__recharge_distribution="lognormal",
        groundwater__recharge_mean=5.0,
        groundwater__recharge_standard_deviation=0.25,
        seed=6,
    )
    ls_prob_lognormal.calculate_landslide_probability()
    np.testing.assert_almost_equal(
        grid_2.at_node["landslide__probability_of_failure"][5], 0.8
    )
    np.testing.assert_almost_equal(
        grid_2.at_node["landslide__probability_of_failure"][9], 0.4
    )


def test_calculate_landslide_probability_lognormal_spatial_method():
    """Testing the main method 'calculate_landslide_probability()' with
    'lognormal_spatial' method.
    """
    grid_3 = RasterModelGrid((5, 4), xy_spacing=(0.2, 0.2))
    gridnum = grid_3.number_of_nodes
    np.random.seed(seed=7)
    grid_3.at_node["topographic__slope"] = np.random.rand(gridnum)
    scatter_dat = np.random.randint(1, 10, gridnum)
    grid_3.at_node["topographic__specific_contributing_area"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_3.at_node["soil__transmissivity"] = np.sort(
        np.random.randint(5, 20, gridnum), -1
    )
    grid_3.at_node["soil__mode_total_cohesion"] = np.sort(
        np.random.randint(30, 900, gridnum)
    )
    grid_3.at_node["soil__minimum_total_cohesion"] = (
        grid_3.at_node["soil__mode_total_cohesion"] - scatter_dat
    )
    grid_3.at_node["soil__maximum_total_cohesion"] = (
        grid_3.at_node["soil__mode_total_cohesion"] + scatter_dat
    )
    grid_3.at_node["soil__internal_friction_angle"] = np.sort(
        np.random.randint(26, 37, gridnum)
    )
    grid_3.at_node["soil__thickness"] = np.sort(np.random.randint(1, 10, gridnum))
    grid_3.at_node["soil__density"] = 2000.0 * np.ones(gridnum)

    ls_prob_lognormal_spatial = LandslideProbability(
        grid_3,
        number_of_iterations=10,
        groundwater__recharge_distribution="lognormal_spatial",
        groundwater__recharge_mean=np.random.randint(2, 7, gridnum),
        groundwater__recharge_standard_deviation=np.random.rand(gridnum),
        seed=7,
    )
    ls_prob_lognormal_spatial.calculate_landslide_probability()
    np.testing.assert_almost_equal(
        grid_3.at_node["landslide__probability_of_failure"][5], 0.4
    )
    np.testing.assert_almost_equal(
        grid_3.at_node["landslide__probability_of_failure"][9], 0.29999999
    )



#! /usr/env/python
r"""
Landlab's Continuous-Time Stochastic (CTS) cellular automata modeling package.

Overview
--------

A CellLab CTS model implements a particular type of cellular
automaton (CA): a continuous-time stochastic CA. The approach is based on that
of Narteau et al. (2002, 2009) and Rozier and Narteau (2014). Like a normal
CA, the domain consists of a lattice of cells, each of which has a discrete
state. Unlike a conventional CA, the updating process is stochastic, and takes
place in continuous rather than discrete time. Any given pair (or "doublet")
of adjacent cell states has a certain specified probability of transition to a
different pair of states. The transition probability is given in the form of an
average *transition rate*, :math:`\lambda` (with dimensions of 1/T); the actual
time of transition is a random variable drawn from an exponential probability
distribution with mean :math:`1/\lambda`.

Subclasses
----------

Landlab provides for several different lattice and connection types:

-  RasterCTS: regular raster grid with transitions between horizontal and
   vertical cell pairs
-  OrientedRasterCTS: like a RasterLCA, but different transition rates can
   be assigned to vertical and horizontal pairs. This property of
   orientation can be used, for example, to implement rules representing
   gravitational attraction, or flow of a fluid with a particular
   direction.
-  RasterD8CTS: like a RasterLCA, but includes diagonal as well as vertical
   and horizontal cell pairs.
-  OrientedRasterD8CTS: as above but orientation also matters.
-  HexCTS: hexagonal grid
-  OrientedHexCTS: hexagonal grid, with transition rates allowed to vary
   according to orientation.

Encoding of "states"
--------------------
As in any traditional cellular automaton model, a LandlabCellularAutomaton
contains a grid of cells ("nodes" in Landlab parlance), each of which is has a
discrete state. States are represented by integers (0, 1, ... N).

In addition, every active link has an *orientation code* and a *link state
code*. The orientation code represents the orientation of the link in space: is
it "vertical" (aligned with the y axis), "horizontal" (aligned with x), or in
some other orientation? The number of possible orientations depends on the
subclass. The base class has only one orientation code (0) (meaning
"orientation doesn't matter), but this is overridden in some of the subclasses.
For example, the OrientedRasterLCA has two orientation codes (0 and 1, for
vertical and horizontal), while the OrientedHexLCA has three (representing the
three axes in a hex-cell / triagonal grid).

Each active link also has a *link state code*. The *state* of a link refers to
its particular combination of nodes and its orientation. For example, link
state 1 refers to a link in which the tail-node has state 0, the head-node has
state 1, and the orientation code is 0. The number of possible link states is
equal to R N^2, where R is the number of orientations (1 to 3, depending on the
subclass) and N is the number of possible node states. The simplest possible
Landlab CA model would have just one orientation code and two possible cell
states, so that there are four unique link states. These would be represented
by the tuples of (tail-node state, head-node state, orientation) as follows::

    link state 0 = (0, 0, 0)
    link state 1 = (0, 1, 0)
    link state 2 = (1, 0, 0)
    link state 3 = (1, 1, 0)

Main data structures
--------------------
node_state : 1d array of int (x number of nodes in grid)
    Node-based grid of node-state codes. This is the grid of cell (sic) states.

link_state_dict : dictionary
    Keys are 3-element tuples that represent the cell-state pairs and
    orientation code for each possible link type; values are the corresponding
    link-state codes. Allows you to look up the link-state code corresponding
    to a particular pair of adjacent nodes with a particular orientation.

node_pair : list (x number of possible link states)
    List of 3-element tuples representing all the various link states. Allows
    you to look up the node states and orientation corresponding to a
    particular link-state ID.

priority_queue : PriorityQueue object containing event records
    Queue containing all future transition events, sorted by time of occurrence
    (from soonest to latest).

next_update : 1d array (x number of links)
    Time (in the future) at which the link will undergo its next transition.
    You might notice that the update time for every scheduled transition is
    also stored with each event in the event queue. Why store it twice?
    Because a scheduled event might be invalidated after the event has been
    scheduled (because another transition has changed one of a link's two
    nodes, for example). The way to tell whether a scheduled event is still
    valid is to compare its time with the corresponding transition time in the
    *next_update* array. If they are different, the event is discarded.

link_orientation : 1d array of int8 (x number of links)
    Orientation code for each link.

link_state : 1d array of int (x number of links)
    State code for each link.

n_trn : 1d array of int (x number of possible link states)
    Number of transitions ("trn" stands for "transition") from a given link
    state.

trn_to : 1d array of ints (x # transitions)
    Stores the link-state code(s) to which a particular transition ID can
    transition.

trn_rate : 1d array of floats (# transitions)
    Rate associated with each link-state transition.


Created GT Sep 2014, starting from link_cap.py.
"""
from __future__ import print_function

import numpy as np
import pylab as plt

import landlab
from landlab.ca.cfuncs import (
    PriorityQueue,
    get_next_event_new,
    push_transitions_to_event_queue,
    run_cts_new,
)

_NEVER = 1e50

_DEBUG = False

_CORE = landlab.grid.base.CORE_NODE


class Transition(object):

    """A transition from one state to another.

    Represents a transition from one state ("from_state") to another
    ("to_state") at a link. The transition probability is represented by a rate
    parameter "rate", with dimensions of 1/T. The probability distribution of
    time until the transition event occurs is exponentional with mean 1/rate.
    The optional name parameter allows the caller to assign a name to any given
    transition.

    Note that from_state and to_state can now be either integer IDs for the
    standardised ordering of the link states (as before), or tuples explicitly
    describing the node state at each end, and the orientation.
    Orientation is 0: horizontal, L-R; 1: vertical, bottom-top.
    For such a tuple, order is (left/bottom, right/top, orientation).

    Transition() constructor sets 3 required properties and 2 optional
    properties for a transition from one cell pair to another.

    Parameters
    ----------
    from_state : int
        Code for the starting state of the cell pair (link)
    to_state : int
        Code for the new state of the cell pair (link)
    rate : float
        Average rate at which this transition occurs (dimension of 1/time)
    name : string (optional)
        Name for this transition
    swap_properties : bool (optional)
        Flag: should properties be exchanged between the two cells?
    """

    def __init__(
        self,
        from_state,
        to_state,
        rate,
        name=None,
        swap_properties=False,
        prop_update_fn=None,
    ):
        """
        Transition() constructor sets 3 required properties and 2 optional
        properties for a transition from one cell pair to another.

        Parameters
        ----------
        from_state : int
            Code for the starting state of the cell pair (link)
        to_state : int
            Code for the new state of the cell pair (link)
        rate : float
            Average rate at which this transition occurs (dimension of 1/time)
        name : string (optional)
            Name for this transition
        swap_properties : bool (optional)
            Flag: should properties be exchanged between the two cells?
        """
        self.from_state = from_state
        self.to_state = to_state
        self.rate = rate
        self.name = name
        self.swap_properties = swap_properties
        self.prop_update_fn = prop_update_fn


class CAPlotter(object):

    """Handle display of a CellLab-CTS grid.

    CAPlotter() constructor keeps a reference to the CA model, and
    optionally a colormap to be used with plots.

    Parameters
    ----------
    ca : LandlabCellularAutomaton object
        Reference to a CA model
    cmap : Matplotlib colormap, optional
        Colormap to be used in plotting

    Examples
    --------
    >>> from landlab import RasterModelGrid, HexModelGrid
    >>> from landlab.ca.celllab_cts import Transition
    >>> from landlab.ca.raster_cts import RasterCTS
    >>> import numpy as np
    >>> grid = RasterModelGrid((3, 5))
    >>> nsd = {0 : 'zero', 1 : 'one'}
    >>> trn_list = []
    >>> trn_list.append(Transition((0, 1, 0), (1, 1, 0), 1.0))
    >>> ins = np.arange(15) % 2
    >>> ca = RasterCTS(grid, nsd, trn_list, ins)
    >>> cap = CAPlotter(ca)
    >>> cap.gridtype
    'rast'
    >>> cap._cmap.name
    'jet'

    >>> from landlab.ca.hex_cts import HexCTS
    >>> import matplotlib
    >>> grid = HexModelGrid(3, 3)
    >>> ins = np.zeros(grid.number_of_nodes, dtype=int)
    >>> ca = HexCTS(grid, nsd, trn_list, ins)
    >>> cap = CAPlotter(ca, cmap=matplotlib.cm.pink)
    >>> cap.gridtype
    'hex'
    >>> cap._cmap.name
    'pink'
    """

    def __init__(self, ca, cmap=None, **kwds):
        """
        CAPlotter() constructor keeps a reference to the CA model, and
        optionally a colormap to be used with plots.

        Parameters
        ----------
        ca : LandlabCellularAutomaton object
            Reference to a CA model
        cmap : Matplotlib colormap, optional
            Colormap to be used in plotting
        """
        import matplotlib

        # Set the colormap; default to matplotlib's "jet" colormap
        if cmap is None:
            self._cmap = matplotlib.cm.jet
        else:
            self._cmap = cmap

        # Keep a reference to the CA model
        self.ca = ca

        # Initialize the plot and remember the grid type
        plt.ion()
        plt.figure(1)
        if type(ca.grid) is landlab.grid.hex.HexModelGrid:
            self.gridtype = "hex"
        else:
            self.gridtype = "rast"

    def update_plot(self):
        """Plot the current node state grid."""
        plt.clf()
        if self.gridtype == "rast":
            nsr = self.ca.grid.node_vector_to_raster(self.ca.node_state)
            plt.imshow(nsr, interpolation="None", origin="lower", cmap=self._cmap)
        else:
            self.ca.grid.hexplot(self.ca.node_state, color_map=self._cmap)

        plt.draw()
        plt.pause(0.001)

    def finalize(self):
        """Wrap up plotting.

        Wrap up plotting by switching off interactive model and showing the
        plot.
        """
        plt.ioff()
        plt.show()


class CellLabCTSModel(object):

    """Link-type (or doublet-type) cellular automaton model.

    A CellLabCTSModel implements a link-type (or doublet-type) cellular
    automaton model. A link connects a pair of cells. Each cell has a state
    (represented by an integer code), and each link also has a state that is
    determined by the states of the cell pair.

    Parameters
    ----------
    model_grid : Landlab ModelGrid object
        Reference to the model's grid
    node_state_dict : dict
        Keys are node-state codes, values are the names associated with
        these codes
    transition_list : list of Transition objects
        List of all possible transitions in the model
    initial_node_states : array of ints (x number of nodes in grid)
        Starting values for node-state grid
    prop_data : array (x number of nodes in grid), optional
        Array of properties associated with each node/cell
    prop_reset_value : number or object, optional
        Default or initial value for a node/cell property (e.g., 0.0).
        Must be same type as *prop_data*.
    """

    def __init__(
        self,
        model_grid,
        node_state_dict,
        transition_list,
        initial_node_states,
        prop_data=None,
        prop_reset_value=None,
        seed=0,
    ):
        """Initialize the CA model.

        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid), optional
            Array of properties associated with each node/cell
        prop_reset_value : number or object, optional
            Default or initial value for a node/cell property (e.g., 0.0).
            Must be same type as *prop_data*.
        seed : int, optional
            Seed for random number generation.
        """

        # Keep a copy of the model grid
        self.grid = model_grid

        # Initialize random number generation
        np.random.seed(seed)

        # Create an array that knows which links are connected to a boundary
        # node
        self.bnd_lnk = np.zeros(self.grid.number_of_links, dtype=np.int8)
        for link_id in range(self.grid.number_of_links):
            if (
                self.grid.status_at_node[self.grid.node_at_link_tail[link_id]] != _CORE
                or self.grid.status_at_node[self.grid.node_at_link_head[link_id]]
                != _CORE
            ):
                self.bnd_lnk[link_id] = True

        # Set up the initial node-state grid
        self.set_node_state_grid(initial_node_states)

        # Current simulation time starts out at zero
        self.current_time = 0.0

        # Figure out how many states there are, and make sure the input data
        # are self consistent.
        #   There are 2 x (N^2) link states, where N is the number of node
        # states. For example, if there are just two node states, 0 and 1, then
        # the possible oriented link pairs are listed below:
        #   0-0 0-1 1-0 1-1  0 0 1 1
        #                    0 1 0 1
        self.num_node_states = len(node_state_dict)
        self.num_node_states_sq = self.num_node_states * self.num_node_states
        self.num_link_states = self.number_of_orientations * self.num_node_states_sq

        assert type(transition_list) is list, "transition_list must be a list!"
        assert transition_list, "Transition list must contain at least one transition"
        last_type = None
        for t in transition_list:
            # TODO: make orientation optional for cases where
            # self.number_of_orientations = 1
            if isinstance(t.from_state, tuple) and isinstance(t.to_state, tuple):
                this_type = tuple
            else:
                this_type = int

            if this_type is tuple:
                # added to allow from and to states to be tuples, not just ids
                for i in t.from_state[:-1]:
                    assert (
                        i < self.num_node_states
                    ), "Transition from_state out of range"
                for i in t.to_state[:-1]:
                    assert i < self.num_node_states, "Transition to_state out of range"
                assert (
                    t.from_state[-1] < self.number_of_orientations
                ), "Encoding for orientation in from_state must be < number of orientations."
                assert (
                    t.to_state[-1] < self.number_of_orientations
                ), "Encoding for orientation in to_state must be < number of orientations."
            else:
                assert (
                    t.from_state < self.num_link_states
                ), "Transition from_state out of range"
                assert (
                    t.to_state < self.num_link_states
                ), "Transition to_state out of range"

            assert (
                last_type == this_type or last_type is None
            ), "All transition types must be either int IDs, or all tuples."
            # this test to ensure all entries are either IDs, or tuples, not
            # mixed
            last_type = this_type

        # Create priority queue for events and next_update array for links
        self.next_update = self.grid.add_zeros("link", "next_update_time")
        self.priority_queue = PriorityQueue()
        self.next_trn_id = -np.ones(self.grid.number_of_links, dtype=np.int)

        # Assign link types from node types
        self.create_link_state_dict_and_pair_list()

        # DEJH adds: convert transition_list to IDs if necessary
        # This is the new part that allows Transition from_ and to_ types
        # to be specified either as ints, or as tuples.
        transition_list_as_ID = transition_list[:]
        if type(transition_list[0].from_state) == tuple:
            # (then they all are..., because of the assertions in __init__)
            for i in range(len(transition_list)):
                transition_list_as_ID[i].from_state = self.link_state_dict[
                    transition_list[i].from_state
                ]
                transition_list_as_ID[i].to_state = self.link_state_dict[
                    transition_list[i].to_state
                ]

        # Set up the information needed to determine the orientation of links
        # in the lattice. The default method just creates an array of zeros
        # (all orientations considered the same), but this will be overridden
        # in subclasses that do use orientation.
        self.setup_array_of_orientation_codes()

        # Using the grid of node states, figure out all the link states
        self.assign_link_states_from_node_types()

        # Create transition data for links
        self.setup_transition_data(transition_list_as_ID)

        # Put the various transitions on the event queue
        self.push_transitions_to_event_queue()

        # In order to keep track of cell "properties", we create an array of
        # indices that refer to locations in the caller's code where properties
        # are tracked.
        self.propid = np.arange(self.grid.number_of_nodes)
        if prop_data is None:
            self.prop_data = np.zeros(self.grid.number_of_nodes)
            self.prop_reset_value = 0.0
        else:
            self.prop_data = prop_data
            self.prop_reset_value = prop_reset_value

    def set_node_state_grid(self, node_states):
        """Set the grid of node-state codes to node_states.

        Sets the grid of node-state codes to node_states. Also checks
        to make sure node_states is in the proper format, which is to
        say, it's a Numpy array of the same length as the number of nodes in
        the grid.

        **Creates**:

        *  self.node_state : 1D array of ints (x number of nodes in grid)
           The node-state array

        Parameters
        ----------
        node_states : 1D array of ints (x number of nodes in grid)

        Notes
        -----
        The node-state array is attached to the grid as a field with the name
        'node_state'.
        """
        assert (
            type(node_states) is np.ndarray
        ), "initial_node_states must be a Numpy array"
        assert (
            len(node_states) == self.grid.number_of_nodes
        ), "length of initial_node_states must equal number of nodes in grid"
        self.grid.at_node["node_state"] = node_states
        self.node_state = node_states

    def create_link_state_dict_and_pair_list(self):
        """Create a dict of link-state to node-state.

        Creates a dictionary that can be used as a lookup table to find out
        which link state corresponds to a particular pair of node states. The
        dictionary keys are 3-element tuples, each of which represents the
        state of the TAIL node, the HEAD node, and the orientation of the link.
        The values are integer codes representing the link state numbers.

        Notes
        -----
        Performance note: making self.node_pair a tuple does not appear to
        change time to lookup values in update_node_states. Changing it to a
        2D array of int actually slows it down.
        """
        self.link_state_dict = {}
        self.node_pair = []
        k = 0
        for orientation in range(self.number_of_orientations):
            for tail_state in range(self.num_node_states):
                for head_state in range(self.num_node_states):
                    self.link_state_dict[(tail_state, head_state, orientation)] = k
                    self.node_pair.append((tail_state, head_state, orientation))
                    k += 1

    def setup_array_of_orientation_codes(self):
        """Create array of active link orientation codes.

        Creates and configures an array that contain the orientation code for
        each active link (and corresponding cell pair).

        **creates**:

        * ``self.link_orientation`` : 1D numpy array

        Notes
        -----

        The setup varies depending on the type of LCA. The default is
        non-oriented, in which case we just have an array of zeros. Subclasses
        will override this method to handle lattices in which orientation
        matters (for example, vertical vs. horizontal in an OrientedRasterLCA).
        """
        self.link_orientation = np.zeros(self.grid.number_of_links, dtype=np.int8)

    def assign_link_states_from_node_types(self):
        """Assign link-state code for each link.

        Takes lists/arrays of "tail" and "head" node IDs for each link, and a
        dictionary that associates pairs of node states (represented as a
        3-element tuple, comprising the TAIL state, FROM state, and
        orientation) to link states.

        **creates**:

        * ``self.link_state`` : 1D numpy array
        """
        self.link_state = np.zeros(self.grid.number_of_links, dtype=int)

        for i in self.grid.active_links:
            orientation = self.link_orientation[i]
            node_pair = (
                self.node_state[self.grid.node_at_link_tail[i]],
                self.node_state[self.grid.node_at_link_head[i]],
                orientation,
            )
            self.link_state[i] = self.link_state_dict[node_pair]

    def setup_transition_data(self, xn_list):
        """Create transition data arrays."""

        # First, create an array that stores the number of possible transitions
        # out of each state.
        n_xn = np.zeros(self.num_link_states, dtype=int)
        for xn in xn_list:
            n_xn[xn.from_state] += 1
        self.n_trn = np.zeros(self.num_link_states, dtype=np.int)

        # Now, create arrays to hold the "to state" and transition rate for each
        # transition. These arrays are dimensioned N x M where N is the number
        # of states, and M is the maximum number of transitions from a single
        # state (for example if state 3 could transition either to state 1 or
        # state 4, and the other states only had one or zero possible
        # transitions, then the maximum would be 2).
        max_transitions = np.max(n_xn)
        self.trn_id = np.zeros((self.num_link_states, max_transitions), dtype=np.int)
        num_transitions = len(xn_list)
        self.trn_to = np.zeros(num_transitions, dtype=np.int)
        self.trn_rate = np.zeros(num_transitions)
        self.trn_propswap = np.zeros(num_transitions, dtype=np.int8)
        self.trn_prop_update_fn = np.zeros(num_transitions, dtype=object)

        for trn in range(num_transitions):
            self.trn_to[trn] = xn_list[trn].to_state
            self.trn_rate[trn] = xn_list[trn].rate
            self.trn_propswap[trn] = xn_list[trn].swap_properties
            if xn_list[trn].prop_update_fn is not None:
                self.trn_prop_update_fn[trn] = xn_list[trn].prop_update_fn
                self._use_propswap_or_callback = True
            from_state = xn_list[trn].from_state
            self.trn_id[from_state, self.n_trn[from_state]] = trn
            self.n_trn[from_state] += 1

    def push_transitions_to_event_queue(self):
        """
        Initializes the event queue by creating transition events for each
        cell pair that has one or more potential transitions and pushing these
        onto the queue. Also records scheduled transition times in the
        self.next_update array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
        >>> ins = np.arange(15) % 2
        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
        >>> ev0 = cts.priority_queue._queue[0]
        >>> np.round(100 * ev0[0])
        12.0
        >>> ev0[2]  # this is the link ID
        16
        >>> ev6 = cts.priority_queue._queue[6]
        >>> np.round(100 * ev6[0])
        27.0
        >>> ev6[2]  # this is the link ID
        6
        >>> cts.next_trn_id[ev0[2]]  # ID of the transition to occur at this link
        3
        >>> cts.next_trn_id[cts.grid.active_links]
        array([-1,  2, -1,  1,  0,  1,  0,  2, -1,  3])
        """
        push_transitions_to_event_queue(
            self.grid.number_of_active_links,
            self.grid.active_links,
            self.n_trn,
            self.link_state,
            self.trn_id,
            self.trn_rate,
            self.next_update,
            self.next_trn_id,
            self.priority_queue,
        )

    def update_link_state_new(self, link, new_link_state, current_time):
        """
        Implements a link transition by updating the current state of the link
        and (if appropriate) choosing the next transition event and pushing it
        on to the event queue.

        Parameters
        ----------
        link : int
            ID of the link to update
        new_link_state : int
            Code for the new state
        current_time : float
            Current time in simulation
        """

        # If the link connects to a boundary, we might have a different state
        # than the one we planned
        if self.bnd_lnk[link]:
            fns = self.node_state[self.grid.node_at_link_tail[link]]
            tns = self.node_state[self.grid.node_at_link_head[link]]
            orientation = self.link_orientation[link]
            new_link_state = int(
                orientation * self.num_node_states_sq + fns * self.num_node_states + tns
            )

        self.link_state[link] = new_link_state
        if self.n_trn[new_link_state] > 0:
            (event_time, trn_id) = get_next_event_new(
                link,
                new_link_state,
                current_time,
                self.n_trn,
                self.trn_id,
                self.trn_rate,
            )
            self.priority_queue.push(link, event_time)
            self.next_update[link] = event_time
            self.next_trn_id[link] = trn_id
        else:
            self.next_update[link] = _NEVER
            self.next_trn_id[link] = -1

    def update_component_data(self, new_node_state_array):
        """Update all component data.

        Call this method to update all data held by the component, if, for
        example, another component or boundary conditions modify the node
        statuses outside the component between run steps.

        This method updates all necessary properties, including both node and
        link states.

        *new_node_state_array* is the updated list of node states, which must
        still all be compatible with the state list originally supplied to
        this component.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.raster_cts import RasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 1, 0), 1.0))
        >>> ins = np.zeros(15, dtype=np.int)
        >>> ca = RasterCTS(grid, nsd, trn_list, ins)
        >>> list(ca.node_state[6:9])
        [0, 0, 0]
        >>> list(ca.link_state[9:13])
        [0, 0, 0, 0]
        >>> len(ca.priority_queue._queue)  # there are no transitions
        0
        >>> nns = np.arange(15) % 2        # make a new node-state grid...
        >>> ca.update_component_data(nns)  # ...and assign it
        >>> list(ca.node_state[6:9])
        [0, 1, 0]
        >>> list(ca.link_state[9:13])
        [2, 1, 2, 1]
        >>> len(ca.priority_queue._queue)  # now there are 5 transitions
        5
        """
        self.set_node_state_grid(new_node_state_array)
        self.assign_link_states_from_node_types()
        self.push_transitions_to_event_queue()

    # @profile
    def run(
        self, run_to, node_state_grid=None, plot_each_transition=False, plotter=None
    ):
        """Run the model forward for a specified period of time.

        Parameters
        ----------
        run_to : float
            Time to run to, starting from self.current_time
        node_state_grid : 1D array of ints (x number of nodes) (optional)
            Node states (if given, replaces model's current node state grid)
        plot_each_transition : bool (optional)
            Option to display the grid after each transition
        plotter : CAPlotter object (optional)
            Needed if caller wants to plot after every transition

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.ca.celllab_cts import Transition
        >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS
        >>> import numpy as np
        >>> grid = RasterModelGrid((3, 5))
        >>> nsd = {0 : 'zero', 1 : 'one'}
        >>> trn_list = []
        >>> trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
        >>> trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
        >>> trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
        >>> ins = np.arange(15) % 2
        >>> cts = OrientedRasterCTS(grid, nsd, trn_list, ins)
        """
        if node_state_grid is not None:
            self.set_node_state_grid(node_state_grid)

        self.current_time = run_cts_new(
            run_to,
            self.current_time,
            self.priority_queue,
            self.next_update,
            self.grid.node_at_link_tail,
            self.grid.node_at_link_head,
            self.node_state,
            self.next_trn_id,
            self.trn_to,
            self.grid.status_at_node,
            self.num_node_states,
            self.num_node_states_sq,
            self.bnd_lnk,
            self.link_orientation,
            self.link_state,
            self.n_trn,
            self.trn_id,
            self.trn_rate,
            self.grid.links_at_node,
            self.grid.active_link_dirs_at_node,
            self.trn_propswap,
            self.propid,
            self.prop_data,
            self.prop_reset_value,
            self.trn_prop_update_fn,
            self,
            plot_each_transition,
            plotter,
        )



import math

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

from landlab import FieldError, HexModelGrid, RasterModelGrid
from landlab.components import FlowAccumulator, FlowDirectorSteepest
from landlab.utils.flow__distance import calculate_flow__distance


def test_no_flow_recievers():
    """Test that correct error is raised when no flow recievers are
    on the grid."""

    # instantiate a model grid, do not run flow accumulation on it

    mg = RasterModelGrid((30, 70))

    # test that the flow distance utility will fail because of a ValueError

    with pytest.raises(FieldError):
        calculate_flow__distance(mg)


def test_no_upstream_array():
    """Test that correct error is raised when no flow__upstream_node_order."""

    # instantiate a model grid, do not run flow accumulation on it

    mg = RasterModelGrid((30, 70))

    # Add a field called topographic__elevation to mg

    mg.add_ones("node", "topographic__elevation")

    # Run the FlowDirectorSteepest component

    fd = FlowDirectorSteepest(mg)
    fd.run_one_step()

    # test that the flow distance utility will fail because of a ValueError

    with pytest.raises(FieldError):
        calculate_flow__distance(mg)


def test_flow__distance_regular_grid_d8():
    """Test to demonstrate that flow__distance utility works as expected with
    regular grids"""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))

    # instantiate an elevation array

    z = np.array(
        [[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0], [0, 0, 0, 0]],
        dtype="float64",
    )

    # add the elevation field to the grid

    mg.add_field("node", "topographic__elevation", z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with D8 algorithm

    flow__distance_expected = np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 0, 0],
            [0, math.sqrt(2), 1, 0],
            [0, 1 + math.sqrt(2), 2, 0],
            [0, 0, 0, 0],
        ],
        dtype="float64",
    )
    flow__distance_expected = np.reshape(
        flow__distance_expected, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # calculating flow directions with FlowAccumulator component

    fr = FlowAccumulator(mg, flow_director="D8")
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True, noclobber=False)
    flow__distance = np.reshape(
        flow__distance, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # test that the flow distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_regular_grid_d4():
    """Test to demonstrate that flow__distance utility works as expected with
    regular grids"""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))

    # instantiate an elevation array

    z = np.array(
        [[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0], [0, 0, 0, 0]],
        dtype="float64",
    )

    # add the elevation field to the grid

    mg.add_field("node", "topographic__elevation", z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with D4 algorithm

    flow__distance_expected = np.array(
        [[0, 0, 0, 0], [0, 1, 0, 0], [0, 2, 1, 0], [0, 3, 2, 0], [0, 0, 0, 0]],
        dtype="float64",
    )
    flow__distance_expected = np.reshape(
        flow__distance_expected, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # calculating flow directions with FlowAccumulator component

    fr = FlowAccumulator(mg, flow_director="D4")
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True, noclobber=False)
    flow__distance = np.reshape(
        flow__distance, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_irregular_grid_d4():
    """Test to demonstrate that flow__distance utility works as expected with irregular grids"""

    # instantiate a model grid

    dx = 1.0
    hmg = HexModelGrid(5, 3, dx)

    # instantiate and add the elevation field

    hmg.add_field(
        "topographic__elevation", hmg.node_x + np.round(hmg.node_y), at="node"
    )

    # instantiate the expected flow__distance array

    flow__distance_expected = np.array(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            dx,
            0.0,
            0.0,
            dx,
            dx,
            2.0 * dx,
            0.0,
            0.0,
            2.0 * dx,
            2.0 * dx,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    # setting boundary conditions

    hmg.set_closed_nodes(hmg.boundary_nodes)

    # calculating flow directions with FlowAccumulator component: D4 algorithm

    fr = FlowAccumulator(hmg, flow_director="D4")
    fr.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(hmg, add_to_grid=True, noclobber=False)

    # test that the flow__distance utility works as expected

    assert_almost_equal(flow__distance_expected, flow__distance, decimal=10)


def test_flow__distance_raster_MFD_diagonals_true():
    """Test of flow__distance utility with a raster grid and MFD."""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))

    # instantiate an elevation array

    z = np.array(
        [[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0], [0, 0, 0, 0]],
        dtype="float64",
    )

    # add the elevation field to the grid

    mg.add_field("node", "topographic__elevation", z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with MFD algorithm

    flow__distance_expected = np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 0, 0],
            [0, math.sqrt(2), 1, 0],
            [0, 1 + math.sqrt(2), 2, 0],
            [0, 0, 0, 0],
        ],
        dtype="float64",
    )
    flow__distance_expected = np.reshape(
        flow__distance_expected, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(
        mg, "topographic__elevation", flow_director="MFD", diagonals=True
    )
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True, noclobber=False)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_raster_MFD_diagonals_false():
    """Test of flow__distance utility with a raster grid and MFD."""

    # instantiate a model grid

    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))

    # instantiate an elevation array

    z = np.array(
        [[0, 0, 0, 0], [0, 21, 10, 0], [0, 31, 20, 0], [0, 32, 30, 0], [0, 0, 0, 0]],
        dtype="float64",
    )

    # add the elevation field to the grid

    mg.add_field("node", "topographic__elevation", z)

    # instantiate the expected flow__distance array
    # considering flow directions calculated with MFD algorithm

    flow__distance_expected = np.array(
        [[0, 0, 0, 0], [0, 1, 0, 0], [0, 2, 1, 0], [0, 3, 2, 0], [0, 0, 0, 0]],
        dtype="float64",
    )
    flow__distance_expected = np.reshape(
        flow__distance_expected, mg.number_of_node_rows * mg.number_of_node_columns
    )

    # setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(
        mg, "topographic__elevation", flow_director="MFD", diagonals=False
    )
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(mg, add_to_grid=True, noclobber=False)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)


def test_flow__distance_raster_D_infinity():
    """Test of flow__distance utility with a raster grid and D infinity."""

    mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))

    # instantiate an elevation array

    z = mg.x_of_node + 3.0 * mg.y_of_node

    # add the elevation field to the grid

    mg.add_field("node", "topographic__elevation", z)

    # instantiate the expected flow_length array

    flow__distance_expected = np.array(
        [
            [0, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 1, 0 + math.sqrt(2.0), 0],
            [0, 2, 1 + math.sqrt(2.0), 0],
            [0, 0, 0, 0],
        ],
        dtype="float64",
    )

    # setting boundary conditions

    mg.set_closed_boundaries_at_grid_edges(
        bottom_is_closed=True,
        left_is_closed=True,
        right_is_closed=True,
        top_is_closed=True,
    )

    # calculating flow directions with FlowAccumulator component

    fa = FlowAccumulator(mg, "topographic__elevation", flow_director="DINF")
    fa.run_one_step()

    # calculating flow distance map

    flow__distance = calculate_flow__distance(
        mg, add_to_grid=True, noclobber=False
    ).reshape(mg.shape)

    # test that the flow__distance utility works as expected

    assert_array_equal(flow__distance_expected, flow__distance)



import os

from ..core.messages import deprecation_message


class DeprecatedSignature(DeprecationWarning):

    msg = "You are using a deprecated calling signature."

    def __init__(self, name, old=None, new=None):
        self._name = name
        self._old = old
        self._new = new

        if old:
            self._old = self._construct_call(name, self._old[0], self._old[1])
        if new:
            self._new = self._construct_call(name, self._new[0], self._new[1])

    @staticmethod
    def _construct_call(name, args, kwds):
        signature = ", ".join(
            [repr(arg) for arg in args]
            + ["{k}={v}".format(k=k, v=repr(v)) for k, v in kwds.items()]
        )
        return "{name}({signature})".format(name=name, signature=signature)

    def __str__(self):
        if self._new:
            use = ">>> grid = {call}".format(call=self._new)
        else:
            use = None

        return os.linesep + deprecation_message(self.msg, use=use)


class DeprecatedRowsColsArguments(DeprecatedSignature):

    msg = """
Passing num_rows, num_cols as separate arguments
is deprecated. Instead, pass a tuple of (num_rows, num_cols)
as the first argument.
"""


class DeprecatedRowsColsKeywords(DeprecatedSignature):
    msg = """
Using num_rows, num_cols keywords is deprecated,
pass a tuple of (num_rows, num_cols) as first argument.
"""


class DeprecatedDxKeyword(DeprecatedSignature):
    msg = "dx keyword is deprecated, use xy_spacing"


class DeprecatedSpacingKeyword(DeprecatedSignature):
    msg = "spacing keyword is deprecated, use xy_spacing"


class DeprecatedSpacingArgument(DeprecatedSignature):
    msg = "spacing keyword is argument, use xy_spacing"


class DeprecatedOriginKeyword(DeprecatedSignature):
    msg = "origin keyword is deprecated, use xy_of_lower_left"



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_boundary_node():
    rmg = RasterModelGrid((5, 6))
    assert rmg.node_has_boundary_neighbor(0)
    assert not rmg.node_has_boundary_neighbor(14)


def test_last_index():
    rmg = RasterModelGrid((4, 5))
    assert rmg.node_has_boundary_neighbor(-1)


def test_id_as_list():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_has_boundary_neighbor([-1, 0]), np.array([True, True]))


def test_id_as_array():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_has_boundary_neighbor(np.arange(20)),
        np.array(
            [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ]
        ),
    )


def test_id_as_array_with_one_interior():
    rmg = RasterModelGrid((5, 5))
    assert_array_equal(
        rmg.node_has_boundary_neighbor(np.arange(25)),
        np.array(
            [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ]
        ),
    )



import numpy as np
import pytest

from landlab.grid import raster_funcs as rfuncs


def test_best_fit_with_bad_args():
    """Raise an error if input arrays are not the same length."""
    len_2 = np.array([0, 1])
    len_3 = np.array([0, 0, 1])

    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_2, len_3, len_3)
    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_3, len_2, len_3)
    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_3, len_3, len_2)


def test_best_fit_plane_in_xy():
    """Best fit plane is the xy-plane."""
    (x, y, z) = (
        np.array([0.0, 1.0, 1.0]),
        np.array([0.0, 0.0, 1.0]),
        np.array([0.0, 0.0, 0.0]),
    )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 0.0
    assert aspect == 90.0

    (x, y, z) = (
        np.array([0.0, 1.0, 1.0, 0.0]),
        np.array([0.0, 0.0, 1.0, 1.0]),
        np.array([0.0, 0.0, 0.0, 0.0]),
    )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 0.0
    assert aspect == 90.0


def test_best_fit_plane_in_xz():
    """Best fit plane is the xz-plane."""
    (x, y, z) = (
        np.array([0.0, 1.0, 1.0]),
        np.array([0.0, 0.0, 0.0]),
        np.array([0.0, 0.0, 1.0]),
    )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 90.0
    assert aspect == 0.0


def test_best_fit_plane_in_yz():
    """Best fit plane is the yz-plane."""
    (x, y, z) = (
        np.array([0.0, 0.0, 0.0]),
        np.array([0.0, 1.0, 1.0]),
        np.array([0.0, 0.0, 1.0]),
    )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 90.0
    assert aspect == 90.0



# -*- coding: utf-8 -*-
"""
Created on Fri Apr  8 08:32:48 2016

@author: RCGlade
"""

import numpy as np

from landlab import INACTIVE_LINK, Component


class DepthDependentDiffuser(Component):

    """
    This component implements a depth and slope dependent linear diffusion rule
    in the style of Johnstone and Hilley (2014).

    Hillslope sediment flux uses depth dependent component inspired by
    Johnstone and Hilley (2014). The flux :math:`q_s` is given as:

    .. math::

        q_s = D S H^* (1.0 - exp( H / H^*)

    where :math:`D` is is the diffusivity, :math:`S` is the slope, :math:`H` is
    the soil depth on links, and :math:`H^*` is the soil transport decay depth.


    This component will ignore soil thickness located at non-core nodes.

    Parameters
    ----------
    grid: ModelGrid
        Landlab ModelGrid object
    linear_diffusivity: float
        Hillslope diffusivity, m**2/yr
    soil_transport_decay_depth: float
        characteristic transport soil depth, m

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import ExponentialWeatherer
    >>> from landlab.components import DepthDependentDiffuser
    >>> mg = RasterModelGrid((5, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(mg.at_node['soil_production__rate'][mg.core_nodes], 1.)
    True
    >>> DDdiff.soilflux(2.)
    >>> np.allclose(mg.at_node['topographic__elevation'][mg.core_nodes], 0.)
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes], -2.)
    True
    >>> np.allclose(mg.at_node['soil__depth'][mg.core_nodes], 2.)
    True

    Now with a slope:

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()
    >>> BRz += mg.node_x/2.
    >>> soilTh[:] = z - BRz
    >>> expweath = ExponentialWeatherer(mg)
    >>> DDdiff = DepthDependentDiffuser(mg)
    >>> expweath.calc_soil_prod_rate()
    >>> np.allclose(
    ...     mg.at_node['soil_production__rate'][mg.core_nodes],
    ...     np.array([ 0.60653066,  0.36787944,  0.22313016]))
    True
    >>> DDdiff.soilflux(2.)
    >>> np.allclose(
    ...     mg.at_node['topographic__elevation'][mg.core_nodes],
    ...     np.array([ 1.47730244,  2.28949856,  3.17558975]))
    True
    >>> np.allclose(mg.at_node['bedrock__elevation'][mg.core_nodes],
    ...     np.array([-0.71306132,  0.26424112,  1.05373968]))
    True
    >>> np.allclose(mg.at_node['soil__depth'], z - BRz)
    True

    Now, we'll test that changing the transport decay depth behaves as expected.

    >>> mg = RasterModelGrid((3, 5))
    >>> soilTh = mg.add_zeros('node', 'soil__depth')
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> BRz = mg.add_zeros('node', 'bedrock__elevation')
    >>> z += mg.node_x.copy()**0.5
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> DDdiff = DepthDependentDiffuser(mg, soil_transport_decay_depth = 0.1)
    >>> DDdiff.soilflux(1)
    >>> soil_decay_depth_point1 = mg.at_node['topographic__elevation'][mg.core_nodes]
    >>> z[:] = 0
    >>> z += mg.node_x.copy()**0.5
    >>> BRz = z.copy() - 1.0
    >>> soilTh[:] = z - BRz
    >>> DDdiff = DepthDependentDiffuser(mg, soil_transport_decay_depth = 1.0)
    >>> DDdiff.soilflux(1)
    >>> soil_decay_depth_1 = mg.at_node['topographic__elevation'][mg.core_nodes]
    >>> np.greater(soil_decay_depth_1[1], soil_decay_depth_point1[1])
    False


    """

    _name = "DepthDependentDiffuser"

    _input_var_names = (
        "topographic__elevation",
        "soil__depth",
        "soil_production__rate",
    )

    _output_var_names = (
        "soil__flux",
        "topographic__slope",
        "topographic__elevation",
        "bedrock__elevation",
        "soil__depth",
    )

    _var_units = {
        "topographic__elevation": "m",
        "topographic__slope": "m/m",
        "soil__depth": "m",
        "soil__flux": "m^2/yr",
        "soil_production__rate": "m/yr",
        "bedrock__elevation": "m",
    }

    _var_mapping = {
        "topographic__elevation": "node",
        "topographic__slope": "link",
        "soil__depth": "node",
        "soil__flux": "link",
        "soil_production__rate": "node",
        "bedrock__elevation": "node",
    }

    _var_doc = {
        "topographic__elevation": "elevation of the ground surface",
        "topographic__slope": "gradient of the ground surface",
        "soil__depth": "depth of soil/weather bedrock",
        "soil__flux": "flux of soil in direction of link",
        "soil_production__rate": "rate of soil production at nodes",
        "bedrock__elevation": "elevation of the bedrock surface",
    }

    def __init__(self, grid, linear_diffusivity=1.0, soil_transport_decay_depth=1.0):
        """Initialize the DepthDependentDiffuser."""

        # Store grid and parameters
        self._grid = grid
        self.K = linear_diffusivity
        self.soil_transport_decay_depth = soil_transport_decay_depth

        # create fields
        # elevation
        if "topographic__elevation" in self.grid.at_node:
            self.elev = self.grid.at_node["topographic__elevation"]
        else:
            self.elev = self.grid.add_zeros("node", "topographic__elevation")

        # slope
        if "topographic__slope" in self.grid.at_link:
            self.slope = self.grid.at_link["topographic__slope"]
        else:
            self.slope = self.grid.add_zeros("link", "topographic__slope")

        # soil depth
        if "soil__depth" in self.grid.at_node:
            self.depth = self.grid.at_node["soil__depth"]
        else:
            self.depth = self.grid.add_zeros("node", "soil__depth")

        # soil flux
        if "soil__flux" in self.grid.at_link:
            self.flux = self.grid.at_link["soil__flux"]
        else:
            self.flux = self.grid.add_zeros("link", "soil__flux")

        # weathering rate
        if "soil_production__rate" in self.grid.at_node:
            self.soil_prod_rate = self.grid.at_node["soil_production__rate"]
        else:
            self.soil_prod_rate = self.grid.add_zeros("node", "soil_production__rate")

        # bedrock elevation
        if "bedrock__elevation" in self.grid.at_node:
            self.bedrock = self.grid.at_node["bedrock__elevation"]
        else:
            self.bedrock = self.grid.add_zeros("node", "bedrock__elevation")

    def soilflux(self, dt):
        """Calculate soil flux for a time period 'dt'.

        Parameters
        ----------

        dt: float (time)
            The imposed timestep.
        """

        # update soil thickness
        self.grid.at_node["soil__depth"][:] = (
            self.grid.at_node["topographic__elevation"]
            - self.grid.at_node["bedrock__elevation"]
        )

        # Calculate soil depth at links.
        H_link = self.grid.map_value_at_max_node_to_link(
            "topographic__elevation", "soil__depth"
        )

        # Calculate gradients
        slope = self.grid.calc_grad_at_link(self.elev)
        slope[self.grid.status_at_link == INACTIVE_LINK] = 0.0

        # Calculate flux
        self.flux[:] = (
            -self.K
            * slope
            * self.soil_transport_decay_depth
            * (1.0 - np.exp(-H_link / self.soil_transport_decay_depth))
        )

        # Calculate flux divergence
        dqdx = self.grid.calc_flux_div_at_node(self.flux)

        # Calculate change in soil depth
        dhdt = self.soil_prod_rate - dqdx

        # Calculate soil depth at nodes
        self.depth[self.grid.core_nodes] += dhdt[self.grid.core_nodes] * dt

        # prevent negative soil thickness
        self.depth[self.depth < 0.0] = 0.0

        # Calculate bedrock elevation
        self.bedrock[self.grid.core_nodes] -= (
            self.soil_prod_rate[self.grid.core_nodes] * dt
        )

        # Update topography
        self.elev[self.grid.core_nodes] = (
            self.depth[self.grid.core_nodes] + self.bedrock[self.grid.core_nodes]
        )

    def run_one_step(self, dt):
        """

        Parameters
        ----------
        dt: float (time)
            The imposed timestep.
        """

        self.soilflux(dt)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import (
    CLOSED_BOUNDARY as CB,
    FIXED_GRADIENT_BOUNDARY as FG,
    FIXED_VALUE_BOUNDARY as FV,
    RasterModelGrid,
)


def test_fixed_link_boundaries_at_grid_edges():
    """test setting fixed link boundaries at grid edges"""

    grid = RasterModelGrid((3, 4))
    grid["node"]["topographic__elevation"] = np.zeros(grid.number_of_nodes)
    grid["link"]["topographic__slope"] = np.zeros(grid.number_of_links)
    grid.set_fixed_link_boundaries_at_grid_edges(False, False, False, True)
    assert_array_equal(
        grid.status_at_node, [FG, FG, FG, FG, FV, 0, 0, FV, FV, FV, FV, FV]
    )

    assert_array_equal(
        grid.status_at_link, [4, 4, 4, 4, 2, 2, 4, 0, 0, 0, 4, 0, 0, 4, 4, 4, 4]
    )


def test_nodata_fixed_links():
    """test setting nodata nodes to fixed gradient"""
    grid = RasterModelGrid((3, 4))
    grid["node"]["topographic__elevation"] = np.zeros(grid.number_of_nodes)
    grid["link"]["topographic__slope"] = np.zeros(grid.number_of_links)
    z = grid["node"]["topographic__elevation"]
    z[3:5] = -9999
    grid.set_nodata_nodes_to_fixed_gradient(z, -9999)
    assert_array_equal(
        grid.status_at_node, [FV, FV, FV, FG, FG, 0, 0, FV, FV, FV, FV, FV]
    )

    assert_array_equal(
        grid.status_at_link, [4, 4, 4, 4, 0, 0, 4, 2, 0, 0, 4, 0, 0, 4, 4, 4, 4]
    )


def test_fixed_gradient_and_value_boundary():
    """testing multiple boundary conditions with fixed links"""
    grid = RasterModelGrid((3, 4))
    grid["node"]["topographic__elevation"] = np.zeros(grid.number_of_nodes)
    grid["link"]["topographic__slope"] = np.zeros(grid.number_of_links)

    grid.set_closed_boundaries_at_grid_edges(False, True, False, True)
    grid.status_at_node[4] = FG
    grid.status_at_node[7] = FG
    grid.status_at_node[1] = 1

    assert_array_equal(
        grid.status_at_node, [CB, FV, CB, CB, FG, 0, 0, FG, CB, CB, CB, CB]
    )

    assert_array_equal(
        grid.status_at_link, [4, 4, 4, 4, 0, 4, 4, 2, 0, 2, 4, 4, 4, 4, 4, 4, 4]
    )



#! /usr/bin/env python
"""Functions to calculate flow distance from divide."""
import numpy as np

from landlab import BAD_INDEX_VALUE, FieldError, RasterModelGrid


def calculate_distance_to_divide(
    grid, longest_path=True, add_to_grid=False, noclobber=True
):
    """
    Calculate the along flow distance from drainage divide to point.

    This utility calculates the along flow distance based on the results of
    running flow accumulation on the grid. It will use the connectivity
    used by the FlowAccumulator (e.g. D4, D8, Dinf).

    Parameters
    ----------
    grid : ModelGrid
    longest_path : bool, optional
        Take the longest (or shortest) path to a drainage divide. Default is
        true.
    add_to_grid : boolean, optional
        Flag to indicate if the stream length field should be added to the
        grid. Default is False. The field name used is ``distance_to_divide``.
    noclobber : boolean, optional
        Flag to indicate if adding the field to the grid should not clobber an
        existing field with the same name. Default is True.

    Returns
    -------
    distance_to_divide : float ndarray
        The distance that has to be covered from an imaginary flow, located in
        each node of the grid, to reach the watershed's outlet.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.distance_to_divide import (
    ...     calculate_distance_to_divide)
    >>> mg = RasterModelGrid((5, 4))
    >>> elev = np.array([0.,  0.,  0., 0.,
    ...                  0., 10., 10., 0.,
    ...                  0., 20., 20., 0.,
    ...                  0., 30., 30., 0.,
    ...                  0.,  0.,  0., 0.])
    >>> _ = mg.add_field('node','topographic__elevation', elev)
    >>> mg.set_closed_boundaries_at_grid_edges(
    ...     bottom_is_closed=False,
    ...     left_is_closed=True,
    ...     right_is_closed=True,
    ...     top_is_closed=True)
    >>> fr = FlowAccumulator(mg, flow_director = 'D8')
    >>> fr.run_one_step()
    >>> distance_to_divide = calculate_distance_to_divide(
    ...     mg,
    ...     add_to_grid=True,
    ...     noclobber=False)
    >>> mg.at_node['distance_to_divide']
    array([ 0.,  3.,  3.,  0.,
            0.,  2.,  2.,  0.,
            0.,  1.,  1.,  0.,
            0.,  0.,  0.,  0.,
            0.,  0.,  0.,  0.])

    Now, let's change to MFD the flow_director method, which routes flow to
    multiple nodes.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.distance_to_divide import (
    ...     calculate_distance_to_divide)
    >>> mg = RasterModelGrid((5, 4), xy_spacing=(1, 1))
    >>> elev = np.array([0.,  0.,  0., 0.,
    ...                  0., 10., 10., 0.,
    ...                  0., 20., 20., 0.,
    ...                  0., 30., 30., 0.,
    ...                  0.,  0.,  0., 0.])
    >>> _ = mg.add_field('node','topographic__elevation', elev)
    >>> mg.set_closed_boundaries_at_grid_edges(
    ...     bottom_is_closed=False,
    ...     left_is_closed=True,
    ...     right_is_closed=True,
    ...     top_is_closed=True)
    >>> fr = FlowAccumulator(mg, flow_director = 'MFD')
    >>> fr.run_one_step()
    >>> distance_to_divide = calculate_distance_to_divide(
    ...     mg,
    ...     add_to_grid=True,
    ...     noclobber=False)
    >>> mg.at_node['distance_to_divide']
    array([ 0.,  3.,  3.,  0.,
            0.,  2.,  2.,  0.,
            0.,  1.,  1.,  0.,
            0.,  0.,  0.,  0.,
            0.,  0.,  0.,  0.])

    The distance_to_divide utility can also work on irregular grids. For the
    example we will use a Hexagonal Model Grid, a special type of Voroni Grid
    that has regularly spaced hexagonal cells.

    >>> from landlab import HexModelGrid, FIXED_VALUE_BOUNDARY, CLOSED_BOUNDARY
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.distance_to_divide import (
    ...     calculate_distance_to_divide)
    >>> dx = 1
    >>> hmg = HexModelGrid(5,3, dx)
    >>> _ = hmg.add_field('topographic__elevation',
    ...                   hmg.node_x + np.round(hmg.node_y),
    ...                   at = 'node')
    >>> hmg.status_at_node[hmg.boundary_nodes] = CLOSED_BOUNDARY
    >>> hmg.status_at_node[0] = FIXED_VALUE_BOUNDARY
    >>> fr = FlowAccumulator(hmg, flow_director = 'D4')
    >>> fr.run_one_step()
    >>> distance_to_divide = calculate_distance_to_divide(
    ...     hmg,
    ...     add_to_grid=True,
    ...     noclobber=False)
    >>> hmg.at_node['distance_to_divide']
    array([ 3.,  0.,  0.,
         0.,  2.,  1.,  0.,
       0.,  1.,  1.,  0.,  0.,
         0.,   0.,  0.,  0.,
            0.,  0.,  0.])
    """
    # check that flow__receiver nodes exists
    if "flow__receiver_node" not in grid.at_node:
        raise FieldError(
            "A 'flow__receiver_node' field is required at the "
            "nodes of the input grid."
        )

    if "flow__upstream_node_order" not in grid.at_node:
        raise FieldError(
            "A 'flow__upstream_node_order' field is required at the "
            "nodes of the input grid."
        )

    if "drainage_area" not in grid.at_node:
        raise FieldError(
            "A 'flow__upstream_node_order' field is required at the "
            "nodes of the input grid."
        )

    # get the reciever nodes, depending on if this is to-one, or to-multiple,
    # we'll need to get a different at-node field.
    if grid.at_node["flow__receiver_node"].size != grid.size("node"):
        to_one = False
    else:
        to_one = True

    flow__receiver_node = grid.at_node["flow__receiver_node"]
    drainage_area = grid.at_node["drainage_area"]

    # get the upstream node order
    flow__upstream_node_order = grid.at_node["flow__upstream_node_order"]

    # get downstream flow link lengths, result depends on type of grid.
    if isinstance(grid, RasterModelGrid):
        flow_link_lengths = grid.length_of_d8[
            grid.at_node["flow__link_to_receiver_node"]
        ]
    else:
        flow_link_lengths = grid.length_of_link[
            grid.at_node["flow__link_to_receiver_node"]
        ]

    # create an array that representes the distance to the divide.
    distance_to_divide = np.zeros(grid.nodes.size)

    if not longest_path:
        distance_to_divide[:] = 2 * grid.size("node") * np.max(flow_link_lengths)

    # iterate through the flow__upstream_node_order backwards.
    for node in reversed(flow__upstream_node_order):

        # if drainage are is equal to node cell area, set distance to zeros
        # this should handle the drainage divide cells as boundary cells have
        # their area set to zero.
        if drainage_area[node] == grid.cell_area_at_node[node]:
            distance_to_divide[node] = 0

        # get flow recievers
        reciever = flow__receiver_node[node]

        if to_one:
            # if not processing an outlet node.
            if reciever != node:

                if longest_path:
                    cond = (
                        distance_to_divide[reciever]
                        < distance_to_divide[node] + flow_link_lengths[node]
                    )
                else:
                    cond = (
                        distance_to_divide[reciever]
                        > distance_to_divide[node] + flow_link_lengths[node]
                    )

                if cond:
                    distance_to_divide[reciever] = (
                        distance_to_divide[node] + flow_link_lengths[node]
                    )

        else:
            # non-existant links are coded with -1
            useable_recievers = np.where(reciever != BAD_INDEX_VALUE)[0]

            for idx in range(len(useable_recievers)):
                r = reciever[useable_recievers][idx]
                fll = flow_link_lengths[node][useable_recievers][idx]

                # if not processing an outlet node.
                if r != node:

                    if longest_path:
                        cond = distance_to_divide[r] < distance_to_divide[node] + fll
                    else:
                        cond = distance_to_divide[r] > distance_to_divide[node] + fll

                    if cond:
                        distance_to_divide[r] = distance_to_divide[node] + fll

    # store on the grid
    if add_to_grid:
        grid.add_field(
            "node", "distance_to_divide", distance_to_divide, noclobber=noclobber
        )

    return distance_to_divide



"""
Unit tests for landlab.components.pet.potential_evapotranspiration_field
"""
import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0.0, 0.0))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(pet):
    assert pet.name == "Potential Evapotranspiration"


def test_input_var_names(pet):
    assert pet.input_var_names == ("radiation__ratio_to_flat_surface",)


def test_output_var_names(pet):
    assert sorted(pet.output_var_names) == [
        "radiation__incoming_shortwave_flux",
        "radiation__net_flux",
        "radiation__net_longwave_flux",
        "radiation__net_shortwave_flux",
        "surface__potential_evapotranspiration_rate",
    ]


def test_var_units(pet):
    assert set(pet.input_var_names) | set(pet.output_var_names) == set(
        dict(pet.units).keys()
    )

    assert pet.var_units("radiation__incoming_shortwave_flux") == "W/m^2"
    assert pet.var_units("radiation__net_flux") == "W/m^2"
    assert pet.var_units("radiation__net_longwave_flux") == "W/m^2"
    assert pet.var_units("radiation__net_shortwave_flux") == "W/m^2"
    assert pet.var_units("radiation__ratio_to_flat_surface") == "None"
    assert pet.var_units("surface__potential_evapotranspiration_rate") == "mm"


def test_grid_shape(pet):
    assert pet.grid.number_of_node_rows == _SHAPE[0]
    assert pet.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(pet):
    assert pet.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(pet):
    assert pet.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(pet):
    for name in pet.grid["node"]:
        field = pet.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            pet.grid.number_of_node_rows * pet.grid.number_of_node_columns,
        )

    for name in pet.grid["cell"]:
        field = pet.grid["cell"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            pet.grid.number_of_cell_rows * pet.grid.number_of_cell_columns,
        )

    with pytest.raises(KeyError):
        pet.grid["not_a_var_name"]


def test_field_initialized_to_zero(pet):
    for name in pet.grid["node"]:
        field = pet.grid["node"][name]
        assert_array_almost_equal(field, np.zeros(pet.grid.number_of_nodes))
    for name in pet.grid["cell"]:
        field = pet.grid["cell"][name]
        assert_array_almost_equal(field, np.zeros(pet.grid.number_of_cells))



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from landlab import BAD_INDEX_VALUE as X, RasterModelGrid


@pytest.mark.parametrize("params", [[(3, 4), {}], [(), {"num_rows": 3, "num_cols": 4}]])
def test_parse_parameters_deprecated_shape(params):
    with pytest.deprecated_call():
        args, kwds = RasterModelGrid._parse_parameters(*params)
    assert args == ((3, 4),)
    assert kwds == {}


@pytest.mark.parametrize(
    "params",
    [[((3, 4),), {"dx": 20.0}], [((3, 4),), {"spacing": 20.0}], [(3, 4, 20.0), {}]],
)
def test_parse_parameters_deprecated_spacing(params):
    with pytest.deprecated_call():
        args, kwds = RasterModelGrid._parse_parameters(*params)
    assert args == ((3, 4),)
    assert kwds == {"xy_spacing": 20.0}


def test_parse_parameters_deprecated_xy_of_lower_left():
    with pytest.deprecated_call():
        args, kwds = RasterModelGrid._parse_parameters([(3, 4)], {"origin": (1.0, 2.0)})
    assert args == ((3, 4),)
    assert kwds == {"xy_of_lower_left": (1.0, 2.0)}


@pytest.mark.parametrize(
    "kwds",
    [
        dict(xy_of_lower_left=(3.0, 4.0)),
        dict(xy_spacing=(30.0, 40.0)),
        dict(xy_spacing=(30.0, 40.0), xy_of_lower_left=(3.0, 4.0)),
        dict(xy_spacing=(30.0, 40.0), xy_of_lower_left=(3.0, 4.0), another_kwd=True),
    ],
)
def test_parse_parameters_new_style(kwds):
    args, new_kwds = RasterModelGrid._parse_parameters([(3, 4)], kwds)
    assert args == ((3, 4),)
    assert new_kwds == kwds


def test_init_with_kwds_classic():

    with pytest.deprecated_call():
        grid = RasterModelGrid(num_rows=4, num_cols=5, xy_spacing=1.0)

    assert grid.number_of_node_rows == 4
    assert grid.number_of_node_columns == 5
    assert grid.dy == 1
    assert grid.dx == 1

    with pytest.deprecated_call():
        grid = RasterModelGrid(3, 7, 2)

    assert grid.number_of_node_rows == 3
    assert grid.number_of_node_columns == 7
    assert grid.dy == 2.0
    assert grid.dx == 2.0


def test_init_new_style():
    grid = RasterModelGrid((4, 5), xy_spacing=2)

    assert grid.number_of_node_rows == 4
    assert grid.number_of_node_columns == 5
    assert grid.dy == 2.0
    assert grid.dx == 2.0

    grid = RasterModelGrid((4, 5))

    assert grid.number_of_node_rows == 4
    assert grid.number_of_node_columns == 5
    assert grid.dy == 1.0
    assert grid.dx == 1.0


def test_spacing_is_float():
    grid = RasterModelGrid((4, 5))
    assert grid.dy == 1.0
    assert isinstance(grid.dy, float)
    assert grid.dx == 1.0
    assert isinstance(grid.dx, float)

    grid = RasterModelGrid((4, 5), xy_spacing=2)
    assert grid.dy == 2.0
    assert isinstance(grid.dy, float)
    assert grid.dx == 2.0
    assert isinstance(grid.dx, float)


def test_grid_dimensions():
    """Test extent of grid with unit spacing."""
    rmg = RasterModelGrid((4, 5))
    assert rmg.extent[0] == rmg.number_of_node_rows - 1
    assert rmg.extent[1] == rmg.number_of_node_columns - 1


def test_grid_dimensions_non_unit_spacing():
    """Test extent of grid with non-unit spacing."""
    rmg = RasterModelGrid((4, 5), xy_spacing=2.0)
    assert rmg.extent[0] == 6.0
    assert rmg.extent[1] == 8.0


def test_nodes_around_point():
    rmg = RasterModelGrid((4, 5))
    surrounding_ids = rmg.nodes_around_point(2.1, 1.1)
    assert_array_equal(surrounding_ids, np.array([7, 12, 13, 8]))

    surrounding_ids = rmg.nodes_around_point(2.1, 0.9)
    assert_array_equal(surrounding_ids, np.array([2, 7, 8, 3]))


def test_neighbor_list_with_scalar_arg():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.active_adjacent_nodes_at_node[6], np.array([7, 11, 5, 1]))
    assert_array_equal(rmg.active_adjacent_nodes_at_node[-1], np.array([X, X, X, X]))
    assert_array_equal(rmg.active_adjacent_nodes_at_node[-2], np.array([X, X, X, 13]))


def test_neighbor_list_with_array_arg():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.active_adjacent_nodes_at_node[[6, -1]],
        np.array([[7, 11, 5, 1], [X, X, X, X]]),
    )


def test_neighbor_list_with_no_args():
    rmg = RasterModelGrid((4, 5))
    expected = np.array(
        [
            [X, X, X, X],
            [X, 6, X, X],
            [X, 7, X, X],
            [X, 8, X, X],
            [X, X, X, X],
            [6, X, X, X],
            [7, 11, 5, 1],
            [8, 12, 6, 2],
            [9, 13, 7, 3],
            [X, X, 8, X],
            [11, X, X, X],
            [12, 16, 10, 6],
            [13, 17, 11, 7],
            [14, 18, 12, 8],
            [X, X, 13, X],
            [X, X, X, X],
            [X, X, X, 11],
            [X, X, X, 12],
            [X, X, X, 13],
            [X, X, X, X],
        ]
    )

    assert_array_equal(rmg.active_adjacent_nodes_at_node, expected)


def test_node_x():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_x,
        np.array(
            [
                0.0,
                1.0,
                2.0,
                3.0,
                4.0,
                0.0,
                1.0,
                2.0,
                3.0,
                4.0,
                0.0,
                1.0,
                2.0,
                3.0,
                4.0,
                0.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ]
        ),
    )


def test_node_y():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_y,
        np.array(
            [
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                3.0,
                3.0,
                3.0,
                3.0,
                3.0,
            ]
        ),
    )


def test_node_x_is_immutable():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.node_x[0] = 0


def test_node_y_is_immutable():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.node_y[0] = 0


def test_node_axis_coordinates():
    rmg = RasterModelGrid((4, 5))
    assert rmg.node_axis_coordinates(axis=0).base is rmg.node_y.base
    assert rmg.node_axis_coordinates(axis=1).base is rmg.node_x.base
    assert rmg.node_axis_coordinates(axis=-1).base is rmg.node_x.base
    assert rmg.node_axis_coordinates(axis=-2).base is rmg.node_y.base


def test_diagonal_list():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[6], np.array([12, 10, 0, 2]))
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[-1], np.array([X, X, 13, X]))
    assert_array_equal(
        rmg.diagonal_adjacent_nodes_at_node[[6, -1]],
        np.array([[12, 10, 0, 2], [X, X, 13, X]]),
    )
    assert_array_equal(
        rmg.diagonal_adjacent_nodes_at_node,
        np.array(
            [
                [6, X, X, X],
                [7, 5, X, X],
                [8, 6, X, X],
                [9, 7, X, X],
                [X, 8, X, X],
                [11, X, X, 1],
                [12, 10, 0, 2],
                [13, 11, 1, 3],
                [14, 12, 2, 4],
                [X, 13, 3, X],
                [16, X, X, 6],
                [17, 15, 5, 7],
                [18, 16, 6, 8],
                [19, 17, 7, 9],
                [X, 18, 8, X],
                [X, X, X, 11],
                [X, X, 10, 12],
                [X, X, 11, 13],
                [X, X, 12, 14],
                [X, X, 13, X],
            ]
        ),
    )


def test_diagonal_list_boundary():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.diagonal_adjacent_nodes_at_node[0], np.array([6, X, X, X]))


def test_node_is_core():
    rmg = RasterModelGrid((4, 5))
    for cell_id in [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 16, 17, 18, 19]:
        assert not rmg.node_is_core(cell_id)
    for cell_id in [6, 7, 8, 11, 12, 13]:
        assert rmg.node_is_core(cell_id)


def test_get_interior_cells():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.node_at_core_cell, np.array([6, 7, 8, 11, 12, 13]))


def test_active_links():
    rmg = RasterModelGrid((4, 5))
    assert rmg.number_of_active_links == 17
    assert_array_equal(
        rmg.active_links,
        np.array([5, 6, 7, 9, 10, 11, 12, 14, 15, 16, 18, 19, 20, 21, 23, 24, 25]),
    )


# def test_active_link_fromnode():
#    rmg = RasterModelGrid((4, 5))
#    assert_array_equal(rmg._activelink_fromnode,
#                       np.array([1, 2, 3, 6, 7, 8, 11, 12, 13,
#                                 5, 6, 7, 8, 10, 11, 12, 13]))
#
#
# def test_active_link_tonode():
#    rmg = RasterModelGrid((4, 5))
#    assert_array_equal(rmg._activelink_tonode,
#                       np.array([6, 7, 8, 11, 12, 13, 16, 17, 18,
#                                 6, 7, 8, 9, 11, 12, 13, 14]))


def test__active_links_at_node_scalar_interior():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg._active_links_at_node([6]), np.array([[5, 9, 14, 10]]).T)


def test__active_links_at_node_scalar_boundary():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg._active_links_at_node([1]), np.array([[-1, -1, 5, -1]]).T)


def test_active_node_with_array_arg():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg._active_links_at_node([6, 7]), np.array([[5, 9, 14, 10], [6, 10, 15, 11]]).T
    )


def test__active_links_at_node_with_no_args():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg._active_links_at_node(),
        np.array(
            [
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    5,
                    6,
                    7,
                    -1,
                    -1,
                    14,
                    15,
                    16,
                    -1,
                    -1,
                    23,
                    24,
                    25,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    9,
                    10,
                    11,
                    12,
                    -1,
                    18,
                    19,
                    20,
                    21,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    5,
                    6,
                    7,
                    -1,
                    -1,
                    14,
                    15,
                    16,
                    -1,
                    -1,
                    23,
                    24,
                    25,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
                [
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    9,
                    10,
                    11,
                    12,
                    -1,
                    18,
                    19,
                    20,
                    21,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                    -1,
                ],
            ]
        ),
    )


def test_nodes_at_link():
    """Test nodes_at_link shares data with tail and head."""
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.nodes_at_link[:, 0], rmg.node_at_link_tail)
    assert_array_equal(rmg.nodes_at_link[:, 1], rmg.node_at_link_head)

    assert np.may_share_memory(rmg.nodes_at_link, rmg.node_at_link_tail)
    assert np.may_share_memory(rmg.nodes_at_link, rmg.node_at_link_head)


def test_node_at_link_tail():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_at_link_tail,
        np.array(
            [
                0,
                1,
                2,
                3,
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
            ]
        ),
    )


def test_node_at_link_head():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.node_at_link_head,
        np.array(
            [
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                16,
                17,
                18,
                19,
            ]
        ),
    )


def test_links_at_node_with_scalar_interior():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.links_at_node[6], np.array([10, 14, 9, 5]))


def test_links_at_node_with_scalar_boundary():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.links_at_node[1], np.array([1, 5, 0, -1]))


def test_links_at_node_with_array_arg():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.links_at_node[6:8], np.array([[10, 14, 9, 5], [11, 15, 10, 6]])
    )


def test_links_at_node_with_no_args():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.links_at_node,
        np.array(
            [
                [0, 4, -1, -1],
                [1, 5, 0, -1],
                [2, 6, 1, -1],
                [3, 7, 2, -1],
                [-1, 8, 3, -1],
                [9, 13, -1, 4],
                [10, 14, 9, 5],
                [11, 15, 10, 6],
                [12, 16, 11, 7],
                [-1, 17, 12, 8],
                [18, 22, -1, 13],
                [19, 23, 18, 14],
                [20, 24, 19, 15],
                [21, 25, 20, 16],
                [-1, 26, 21, 17],
                [27, -1, -1, 22],
                [28, -1, 27, 23],
                [29, -1, 28, 24],
                [30, -1, 29, 25],
                [-1, -1, 30, 26],
            ]
        ),
    )


def test_face_at_link():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.face_at_link,
        np.array(
            [
                X,
                X,
                X,
                X,
                X,
                0,
                1,
                2,
                X,
                3,
                4,
                5,
                6,
                X,
                7,
                8,
                9,
                X,
                10,
                11,
                12,
                13,
                X,
                14,
                15,
                16,
                X,
                X,
                X,
                X,
                X,
            ]
        ),
    )


def test_grid_coords_to_node_id_with_scalar():
    rmg = RasterModelGrid((4, 5))
    assert rmg.grid_coords_to_node_id(3, 4) == 19


def test_grid_coords_to_node_id_with_array():
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(rmg.grid_coords_to_node_id((3, 2), (4, 1)), np.array([19, 11]))


def test_grid_coords_to_node_id_outside_of_grid():
    rmg = RasterModelGrid((4, 5))
    with pytest.raises(ValueError):
        rmg.grid_coords_to_node_id(5, 0)


def test_diagonal_adjacent_nodes_at_node():
    """Test diagonally adjacent nodes."""
    rmg = RasterModelGrid((4, 5))
    assert_array_equal(
        rmg.diagonal_adjacent_nodes_at_node,
        np.array(
            [
                [6, X, X, X],
                [7, 5, X, X],
                [8, 6, X, X],
                [9, 7, X, X],
                [X, 8, X, X],
                [11, X, X, 1],
                [12, 10, 0, 2],
                [13, 11, 1, 3],
                [14, 12, 2, 4],
                [X, 13, 3, X],
                [16, X, X, 6],
                [17, 15, 5, 7],
                [18, 16, 6, 8],
                [19, 17, 7, 9],
                [X, 18, 8, X],
                [X, X, X, 11],
                [X, X, 10, 12],
                [X, X, 11, 13],
                [X, X, 12, 14],
                [X, X, 13, X],
            ]
        ),
    )



# -*- coding: utf-8 -*-
"""
Unit tests for source_tracking_algorithm.py
@author: Sai Nudurupati & Erkan Istanbulluoglu
"""

import numpy as np
import pytest

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator
from landlab.utils import find_unique_upstream_hsd_ids_and_fractions, track_source


def test_route_to_multiple_error_raised():
    grid = RasterModelGrid((5, 5), xy_spacing=(1.0, 1.0))
    grid.at_node["topographic__elevation"] = np.array(
        [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            4.0,
            5.0,
            1.0,
            5.0,
            0.0,
            3.0,
            5.0,
            3.0,
            0.0,
            5.0,
            4.0,
            5.0,
            2.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
    )
    grid.status_at_node[10] = 0
    grid.status_at_node[14] = 0
    fa = FlowAccumulator(grid, flow_director="MFD")
    fa.run_one_step()
    hsd_ids = np.empty(grid.number_of_nodes, dtype=int)
    hsd_ids[:] = 1
    hsd_ids[2:5] = 0
    hsd_ids[7:10] = 0

    with pytest.raises(NotImplementedError):
        track_source(grid, hsd_ids)


def test_track_source():
    """Unit tests for track_source().
    """
    grid = RasterModelGrid((5, 5), xy_spacing=(1.0, 1.0))
    grid.at_node["topographic__elevation"] = np.array(
        [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            4.0,
            5.0,
            1.0,
            5.0,
            0.0,
            3.0,
            5.0,
            3.0,
            0.0,
            5.0,
            4.0,
            5.0,
            2.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
    )
    grid.status_at_node[10] = 0
    grid.status_at_node[14] = 0
    fr = FlowAccumulator(grid, flow_director="D8")
    fr.run_one_step()
    r = grid.at_node["flow__receiver_node"]
    assert r[6] == 10
    assert r[7] == 8
    assert r[18] == 14
    hsd_ids = np.empty(grid.number_of_nodes, dtype=int)
    hsd_ids[:] = 1
    hsd_ids[2:5] = 0
    hsd_ids[7:10] = 0
    (hsd_upstr, flow_accum) = track_source(grid, hsd_ids)
    assert hsd_upstr[8] == [1, 0, 0]
    assert hsd_upstr[14] == [1, 1, 1, 1, 0, 0, 1]
    assert flow_accum[14] == 7


def test_find_unique_upstream_hsd_ids_and_fractions():
    """Unit tests find_unique_upstream_hsd_ids_and_fractions().
    """
    grid = RasterModelGrid((5, 5), xy_spacing=(1.0, 1.0))
    grid.at_node["topographic__elevation"] = np.array(
        [
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            4.0,
            5.0,
            1.0,
            5.0,
            0.0,
            3.0,
            5.0,
            3.0,
            0.0,
            5.0,
            4.0,
            5.0,
            2.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
            5.0,
        ]
    )
    grid.status_at_node[10] = 0
    grid.status_at_node[14] = 0
    fr = FlowAccumulator(grid, flow_director="D8")
    fr.run_one_step()
    hsd_ids = np.empty(grid.number_of_nodes, dtype=int)
    hsd_ids[:] = 1
    hsd_ids[2:5] = 0
    hsd_ids[7:10] = 0
    (hsd_upstr, flow_accum) = track_source(grid, hsd_ids)
    (uniq_ids, coeff) = find_unique_upstream_hsd_ids_and_fractions(hsd_upstr)
    np.testing.assert_almost_equal(
        np.sort(np.array(coeff[8])), np.array([0.33333333, 0.66666667])
    )



import pytest

from landlab import RasterModelGrid
from landlab.components import FlowAccumulator, SteepnessFinder


def test_route_to_multiple_error_raised():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        SteepnessFinder(mg)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_lower_left_cell():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.1, 0.1), np.array([0, 3, 4, 1]))


def test_upper_left_cell():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.1, 0.9), np.array([0, 3, 4, 1]))


def test_upper_right_cell():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.9, 0.9), np.array([0, 3, 4, 1]))


def test_lower_right_cell():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.9, 0.1), np.array([0, 3, 4, 1]))


def test_on_left_edge():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.0, 0.5), np.array([0, 3, 4, 1]))


def test_on_bottom_edge():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.5, 0.0), np.array([0, 3, 4, 1]))


def test_on_right_edge():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(1.0, 0.5), np.array([1, 4, 5, 2]))


def test_on_top_edge():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.5, 1.0), np.array([3, 6, 7, 4]))


def test_on_lower_left_corner():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.0, 0.0), np.array([0, 3, 4, 1]))


def test_on_upper_right_corner():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(1.0, 1.0), np.array([4, 7, 8, 5]))


def test_on_upper_left_corner():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(0.0, 1.0), np.array([3, 6, 7, 4]))


def test_on_lower_right_corner():
    rmg = RasterModelGrid((3, 3))
    assert_array_equal(rmg.nodes_around_point(1.0, 0.0), np.array([1, 4, 5, 2]))



from .base import (
    ACTIVE_LINK,
    BAD_INDEX_VALUE,
    CLOSED_BOUNDARY,
    CORE_NODE,
    FIXED_GRADIENT_BOUNDARY,
    FIXED_LINK,
    FIXED_VALUE_BOUNDARY,
    INACTIVE_LINK,
    LOOPED_BOUNDARY,
    ModelGrid,
)
from .create import create_and_initialize_grid, create_grid
from .hex import HexModelGrid
from .network import NetworkModelGrid
from .radial import RadialModelGrid
from .raster import RasterModelGrid
from .voronoi import VoronoiDelaunayGrid

__all__ = [
    "ModelGrid",
    "HexModelGrid",
    "RadialModelGrid",
    "RasterModelGrid",
    "VoronoiDelaunayGrid",
    "NetworkModelGrid",
    "BAD_INDEX_VALUE",
    "CORE_NODE",
    "FIXED_VALUE_BOUNDARY",
    "FIXED_GRADIENT_BOUNDARY",
    "LOOPED_BOUNDARY",
    "CLOSED_BOUNDARY",
    "ACTIVE_LINK",
    "FIXED_LINK",
    "INACTIVE_LINK",
    "create_and_initialize_grid",
    "create_grid",
]



import pytest

from landlab import HexModelGrid, RasterModelGrid
from landlab.components import ChiFinder, FlowAccumulator


def test_route_to_multiple_error_raised():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg, flow_director="MFD")
    fa.run_one_step()

    with pytest.raises(NotImplementedError):
        ChiFinder(mg, min_drainage_area=1.0, reference_concavity=1.0)


def test_bad_reference_area():
    mg = RasterModelGrid((10, 10))
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg)
    fa.run_one_step()

    with pytest.raises(ValueError):
        ChiFinder(mg, min_drainage_area=1.0, reference_area=-1)


def test_functions_with_Hex():
    mg = HexModelGrid(10, 10)
    z = mg.add_zeros("node", "topographic__elevation")
    z += mg.x_of_node + mg.y_of_node
    fa = FlowAccumulator(mg)
    fa.run_one_step()

    ch = ChiFinder(mg, min_drainage_area=1.0, reference_concavity=1.0)
    ch.calculate_chi()



from .sort import (
    reindex_by_xy,
    reorder_links_at_patch,
    sort_graph,
    sort_nodes,
    sort_patches,
)

__all__ = [
    "sort_graph",
    "sort_patches",
    "sort_nodes",
    "reindex_by_xy",
    "reorder_links_at_patch",
]



from ..flow_director import flow_direction_DN
from ..flow_director.flow_direction_DN import flow_directions
from .flow_director_d8 import FlowDirectorD8
from .flow_director_dinf import FlowDirectorDINF
from .flow_director_mfd import FlowDirectorMFD
from .flow_director_steepest import FlowDirectorSteepest

__all__ = [
    "FlowDirectorD8",
    "FlowDirectorSteepest",
    "FlowDirectorMFD",
    "FlowDirectorDINF",
    "flow_directions",
    "flow_direction_DN",
]



import pytest
from numpy.testing import assert_array_equal

from landlab.grid.unstructured.links import (
    in_link_ids_at_node,
    link_ids_at_node,
    out_link_ids_at_node,
)


def test_array_length_mismatch():
    with pytest.raises(ValueError):
        link_ids_at_node(([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7]))


def test_link_arg_as_scalars():
    links, count = link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0, 0])
    assert_array_equal(count, [1, 1])

    links, count = in_link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0])
    assert_array_equal(count, [0, 1])

    links, count = out_link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0])
    assert_array_equal(count, [1, 0])


def test_link_arg_as_pairs():
    links, count = link_ids_at_node([(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2, 0, 1, 2])
    assert_array_equal(count, [3, 1, 1, 1])

    links, count = in_link_ids_at_node([(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2])
    assert_array_equal(count, [0, 1, 1, 1])

    links, count = out_link_ids_at_node([(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2])
    assert_array_equal(count, [3, 0, 0, 0])


def test_link_arg_as_tuple_sequence():
    links, count = link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [1, 0, 0, 1])
    assert_array_equal(count, [2, 2])

    links, count = in_link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [1, 0])
    assert_array_equal(count, [1, 1])

    links, count = out_link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [0, 1])
    assert_array_equal(count, [1, 1])


def test_links_3x3_quad():
    link_ends = [
        (0, 3),
        (1, 4),
        (2, 5),
        (3, 6),
        (4, 7),
        (5, 8),
        (0, 1),
        (1, 2),
        (3, 4),
        (4, 5),
        (6, 7),
        (7, 8),
    ]

    (links, count) = link_ids_at_node(list(zip(*link_ends)), number_of_nodes=9)

    assert_array_equal(
        links,
        [0, 6, 6, 1, 7, 7, 2, 0, 3, 8, 1, 8, 4, 9, 2, 9, 5, 3, 10, 4, 10, 11, 5, 11],
    )
    assert_array_equal(count, [2, 3, 2, 3, 4, 3, 2, 3, 2])


def test_out_links_2x2_quad():
    link_ends = [(0, 2), (1, 3), (0, 1), (2, 3)]
    (links, count) = out_link_ids_at_node(zip(*link_ends), number_of_nodes=4)
    assert_array_equal(links, [0, 2, 1, 3])
    assert_array_equal(count, [2, 1, 1, 0])


def test_out_links_3x3_quad():
    link_ends = [
        (0, 3),
        (1, 4),
        (2, 5),
        (3, 6),
        (4, 7),
        (5, 8),
        (0, 1),
        (1, 2),
        (3, 4),
        (4, 5),
        (6, 7),
        (7, 8),
    ]

    (links, count) = out_link_ids_at_node(link_ends, number_of_nodes=9)
    assert_array_equal(links, [0, 6, 1, 7, 2, 3, 8, 4, 9, 5, 10, 11])
    assert_array_equal(count, [2, 2, 1, 2, 2, 1, 1, 1, 0])



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 10:36:03 2015

@author: gtucker
"""
import warnings

from landlab import HexModelGrid


def test_shape_dep_warning():
    """Test dep warning on use of shape keyword instead of node_layout."""
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("always")
        # Trigger the deprecation warning.
        HexModelGrid(3, 2, shape="rect")
        # Verify some things
        catmsg = ""
        for warn in w:
            catmsg += str(warn.message)
        assert "node_layout" in catmsg
